"""Base class for Module and ConstantsModule."""

import ast
import datetime
import socket
from contextlib import suppress
from keyword import iskeyword
from pathlib import Path
from types import CodeType
from typing import Dict, List, Optional, Sequence, Set, Tuple, Union

from ._compat import importlib_metadata
from .common import TemporaryPath
from .exception import ConfigError

__all__ = ["ConstantsModule", "Module"]


class DistributionCache(importlib_metadata.PathDistribution):
    """Cache the distribution package."""

    _cachedir = TemporaryPath()

    @staticmethod
    def at(path: Union[str, Path]):
        return DistributionCache(Path(path))

    at.__doc__ = importlib_metadata.PathDistribution.at.__doc__

    @classmethod
    def from_name(cls, name: str):
        distribution = super().from_name(name)

        # Cache dist-info files in a temporary directory
        normalized_name = getattr(distribution, "_normalized_name", None)
        if normalized_name is None:
            normalized_name = importlib_metadata.Prepared.normalize(name)
        source_path = getattr(distribution, "_path", None)
        if source_path is None:
            mask = f"{normalized_name}-{distribution.version}*-info"
            source_path = list(distribution.locate_file("").glob(mask))[0]
        if not source_path.exists():
            raise importlib_metadata.PackageNotFoundError(name)

        target_name = f"{normalized_name}-{distribution.version}.dist-info"
        target_path = cls._cachedir.path / target_name
        target_path.mkdir(exist_ok=True)

        purelib = None
        if source_path.name.endswith(".dist-info"):
            for source in source_path.rglob("*"):  # type: Path
                target = target_path / source.relative_to(source_path)
                if source.is_dir():
                    target.mkdir(exist_ok=True)
                else:
                    target.write_bytes(source.read_bytes())
        elif source_path.is_file():
            # old egg-info file is converted to dist-info
            target = target_path / "METADATA"
            target.write_bytes(source_path.read_bytes())
            purelib = (source_path.parent / (normalized_name + ".py")).exists()
        else:
            # Copy minimal data from egg-info directory into dist-info
            source = source_path / "PKG-INFO"
            if source.is_file():
                target = target_path / "METADATA"
                target.write_bytes(source.read_bytes())
            source = source_path / "top_level.txt"
            if source.is_file():
                target = target_path / "top_level.txt"
                target.write_bytes(source.read_bytes())
            purelib = not source_path.joinpath("not-zip-safe").is_file()

        cls._write_wheel_distinfo(target_path, purelib)
        cls._write_record_distinfo(target_path)

        return cls.at(target_path)

    from_name.__doc__ = importlib_metadata.PathDistribution.from_name.__doc__

    @staticmethod
    def _write_wheel_distinfo(target_path: Path, purelib: bool):
        """Create a WHEEL file if it doesn't exist"""
        target = target_path / "WHEEL"
        if not target.exists():
            project = Path(__file__).parent.name
            version = importlib_metadata.version(project)
            root_is_purelib = "true" if purelib else "false"
            text = [
                "Wheel-Version: 1.0",
                f"Generator: {project} ({version})",
                f"Root-Is-Purelib: {root_is_purelib}",
                "Tag: py3-none-any",
            ]
            target.write_text("\n".join(text))

    @staticmethod
    def _write_record_distinfo(target_path: Path):
        """Recreate a minimal RECORD file"""
        target_name = target_path.name
        record = []
        for file in target_path.iterdir():
            record.append(f"{target_name}/{file.name},,")
        record.append(f"{target_name}/RECORD,,")
        target = target_path / "RECORD"
        target.write_text("\n".join(record), encoding="utf-8")


class Module:
    """
    The Module class.
    """

    def __init__(
        self,
        name: str,
        path: Optional[Sequence[Union[Path, str]]] = None,
        file_name: Optional[Union[Path, str]] = None,
        parent: Optional["Module"] = None,
    ):
        self.name: str = name
        self.path: Optional[List[Path]] = None
        if path:
            self.path = [Path(p) for p in path]
        self.file = file_name
        self.parent: Optional["Module"] = parent
        self.code: Optional[CodeType] = None
        self.distribution: Optional[DistributionCache] = None
        self.exclude_names: Set[str] = set()
        self.global_names: Set[str] = set()
        self.ignore_names: Set[str] = set()
        self.in_import: bool = True
        self.source_is_string: bool = False
        self.source_is_zip_file: bool = False
        self.in_file_system = 1
        # cache the dist-info files (metadata)
        self.update_distribution(name)

    @property
    def file(self) -> Optional[Path]:
        """Module filename"""
        return self._file

    @file.setter
    def file(self, file_name: Optional[Union[Path, str]]):
        self._file: Optional[Path] = Path(file_name) if file_name else None

    def update_distribution(self, name: str) -> None:
        """Update the distribution cache based on its name.
        This method may be used to link an distribution's name to a module.

        Example: ModuleFinder cannot detects the distribution of _cffi_backend
        but in a hook we can link it to 'cffi'.
        """
        try:
            distribution = DistributionCache.from_name(name)
        except importlib_metadata.PackageNotFoundError:
            distribution = None
        if distribution is None:
            return
        try:
            requires = importlib_metadata.requires(distribution.name) or []
        except importlib_metadata.PackageNotFoundError:
            requires = []
        for req in requires:
            req_name = req.partition(" ")[0]
            with suppress(importlib_metadata.PackageNotFoundError):
                DistributionCache.from_name(req_name)
        self.distribution = distribution

    def __repr__(self) -> str:
        parts = [f"name={self.name!r}"]
        if self.file is not None:
            parts.append(f"file={self.file!r}")
        if self.path is not None:
            parts.append(f"path={self.path!r}")
        join_parts = ", ".join(parts)
        return f"<Module {join_parts}>"

    @property
    def in_file_system(self) -> int:
        """Returns a value indicating where the module/package will be stored:
        0. in a zip file (not directly in the file system)
        1. in the file system, package with modules and data
        2. in the file system, only detected modules."""
        if self.parent is not None:
            return self.parent.in_file_system
        if self.path is None or self.file is None:
            return 0
        return self._in_file_system

    @in_file_system.setter
    def in_file_system(self, value: int) -> None:
        self._in_file_system: int = value


class ConstantsModule:
    """
    Base ConstantsModule class.
    """

    def __init__(
        self,
        release_string: Optional[str] = None,
        copyright_string: Optional[str] = None,
        module_name: str = "BUILD_CONSTANTS",
        time_format: str = "%B %d, %Y %H:%M:%S",
        constants: Optional[List[str]] = None,
    ):
        self.module_name: str = module_name
        self.time_format: str = time_format
        self.values: Dict[str, str] = {}
        self.values["BUILD_RELEASE_STRING"] = release_string
        self.values["BUILD_COPYRIGHT"] = copyright_string
        if constants:
            for constant in constants:
                parts = constant.split("=", maxsplit=1)
                if len(parts) == 1:
                    name = constant
                    value = None
                else:
                    name, string_value = parts
                    value = ast.literal_eval(string_value)
                if (not name.isidentifier()) or iskeyword(name):
                    raise ConfigError(
                        f"Invalid constant name in ConstantsModule ({name!r})"
                    )
                self.values[name] = value
        self.module_path: TemporaryPath = TemporaryPath("constants.py")

    def create(self, modules: List[Module]) -> Tuple[Path, str]:
        """
        Create the module which consists of declaration statements for each
        of the values.
        """
        today = datetime.datetime.today()
        source_timestamp = 0
        for module in modules:
            if module.file is None or module.source_is_string:
                continue
            if module.source_is_zip_file:
                continue
            if not module.file.exists():
                raise ConfigError(
                    f"No file named {module.file!s} (for module {module.name})"
                )
            timestamp = module.file.stat().st_mtime
            source_timestamp = max(source_timestamp, timestamp)
        stamp = datetime.datetime.fromtimestamp(source_timestamp)
        self.values["BUILD_TIMESTAMP"] = today.strftime(self.time_format)
        self.values["BUILD_HOST"] = socket.gethostname().split(".")[0]
        self.values["SOURCE_TIMESTAMP"] = stamp.strftime(self.time_format)
        parts = []
        names = list(self.values.keys())
        names.sort()
        for name in names:
            value = self.values[name]
            parts.append(f"{name} = {value!r}")
        self.module_path.path.write_text("\n".join(parts), encoding="utf-8")
        return self.module_path.path, self.module_name



"""
DLL list of MSVC runtimes.

Extracted from:
    https://github.com/conda-forge/vc-feedstock/blob/master/recipe/meta.yaml

"""

FILES = (
    # VC 2015 and 2017
    "api-ms-win-core-console-l1-1-0.dll",
    "api-ms-win-core-datetime-l1-1-0.dll",
    "api-ms-win-core-debug-l1-1-0.dll",
    "api-ms-win-core-errorhandling-l1-1-0.dll",
    "api-ms-win-core-file-l1-1-0.dll",
    "api-ms-win-core-file-l1-2-0.dll",
    "api-ms-win-core-file-l2-1-0.dll",
    "api-ms-win-core-handle-l1-1-0.dll",
    "api-ms-win-core-heap-l1-1-0.dll",
    "api-ms-win-core-interlocked-l1-1-0.dll",
    "api-ms-win-core-libraryloader-l1-1-0.dll",
    "api-ms-win-core-localization-l1-2-0.dll",
    "api-ms-win-core-memory-l1-1-0.dll",
    "api-ms-win-core-namedpipe-l1-1-0.dll",
    "api-ms-win-core-processenvironment-l1-1-0.dll",
    "api-ms-win-core-processthreads-l1-1-0.dll",
    "api-ms-win-core-processthreads-l1-1-1.dll",
    "api-ms-win-core-profile-l1-1-0.dll",
    "api-ms-win-core-rtlsupport-l1-1-0.dll",
    "api-ms-win-core-string-l1-1-0.dll",
    "api-ms-win-core-synch-l1-1-0.dll",
    "api-ms-win-core-synch-l1-2-0.dll",
    "api-ms-win-core-sysinfo-l1-1-0.dll",
    "api-ms-win-core-timezone-l1-1-0.dll",
    "api-ms-win-core-util-l1-1-0.dll",
    "api-ms-win-crt-conio-l1-1-0.dll",
    "api-ms-win-crt-convert-l1-1-0.dll",
    "api-ms-win-crt-environment-l1-1-0.dll",
    "api-ms-win-crt-filesystem-l1-1-0.dll",
    "api-ms-win-crt-heap-l1-1-0.dll",
    "api-ms-win-crt-locale-l1-1-0.dll",
    "api-ms-win-crt-math-l1-1-0.dll",
    "api-ms-win-crt-multibyte-l1-1-0.dll",
    "api-ms-win-crt-private-l1-1-0.dll",
    "api-ms-win-crt-process-l1-1-0.dll",
    "api-ms-win-crt-runtime-l1-1-0.dll",
    "api-ms-win-crt-stdio-l1-1-0.dll",
    "api-ms-win-crt-string-l1-1-0.dll",
    "api-ms-win-crt-time-l1-1-0.dll",
    "api-ms-win-crt-utility-l1-1-0.dll",
    "concrt140.dll",
    "msvcp140_1.dll",
    "msvcp140_2.dll",
    "msvcp140.dll",
    "ucrtbase.dll",
    "vcamp140.dll",
    "vccorlib140.dll",
    "vcomp140.dll",
    "vcruntime140.dll",
    # VS 2019
    "vcruntime140_1.dll",
    "msvcp140_atomic_wait.dll",
    "msvcp140_codecvt_ids.dll",
)

FILES_TO_DUPLICATE = ("vcruntime140.dll",)



"""Implements `Parser` interface to create an abstraction to parse binary
files."""

import os
import re
import shutil
import stat
import sys
from abc import ABC, abstractmethod
from contextlib import suppress
from pathlib import Path
from subprocess import PIPE, CalledProcessError, check_call, check_output, run
from typing import Any, Dict, List, Set, Union

from .common import TemporaryPath

try:
    import lief
except ImportError:
    lief = None
else:
    lief.logging.set_level(lief.logging.LOGGING_LEVEL.ERROR)

try:
    from cx_Freeze import util
except ImportError:
    util = None

PE_EXT = (".exe", ".dll", ".pyd")
MAGIC_ELF = b"\x7fELF"
NON_ELF_EXT = ".a:.c:.h:.py:.pyc:.pyi:.pyx:.pxd:.txt:.html:.xml".split(":")
NON_ELF_EXT += ".png:.jpg:.gif:.jar:.json".split(":")

# The default is to use lief if it is installed.
# To disable the experimental feature in Windows:
# set CX_FREEZE_BIND=imagehlp
LIEF_ENABLED = os.environ.get("CX_FREEZE_BIND", "lief") == "lief" and lief


class Parser(ABC):
    """`Parser` interface."""

    def __init__(self, silent: int = 0) -> None:
        self.dependent_files: Dict[Path, Set[Path]] = {}
        self._silent: int = silent

    @abstractmethod
    def get_dependent_files(self, path: Union[str, Path]) -> Set[Path]:
        """Return the file's dependencies using platform-specific tools (the
        imagehlp library on Windows, otool on Mac OS X and ldd on Linux);
        limit this list by the exclusion lists as needed.
        (Implemented separately for each platform.)"""


class PEParser(Parser):
    """`PEParser` is based on the `lief` package. The use of LIEF is
    experimental. If it is not installed or disabled use the old friend
    `cx_Freeze.util` extension module."""

    @staticmethod
    def is_pe(path: Union[str, Path]) -> bool:
        """Determines whether the file is a PE file."""
        if isinstance(path, str):
            path = Path(path)
        return path.suffix.lower().endswith(PE_EXT) and path.is_file()

    def _get_dependent_files_lief(self, path: Path) -> Set[Path]:
        with path.open("rb", buffering=0) as raw:
            binary = lief.PE.parse(raw, path.name)
        if not binary:
            return set()

        libraries: List[str] = []
        if binary.has_imports:
            libraries += binary.libraries
        if getattr(binary, "has_delay_imports", False):
            for delay_import in binary.delay_imports:
                libraries.append(delay_import.name)

        dependent_files: Set[Path] = set()
        orig_path: List[str] = os.environ["PATH"]
        search_path: List[str] = sys.path + orig_path.split(os.pathsep)
        for library_name in libraries:
            for directory in search_path:
                library_path = Path(directory, library_name)
                if library_path.is_file():
                    dependent_files.add(library_path)
                    break
        return dependent_files

    def _get_dependent_files_imagehlp(self, path: Path) -> Set[Path]:
        dependent_files: Set[Path] = set()
        orig_path = os.environ["PATH"]
        os.environ["PATH"] = os.pathsep.join(sys.path) + os.pathsep + orig_path
        try:
            files: List[str] = util.GetDependentFiles(path)
        except util.BindError as exc:
            # Sometimes this gets called when path is not actually
            # a library (See issue 88).
            if self._silent < 3:
                print("WARNING: ignoring error during ", end="")
                print(f"GetDependentFiles({path}):", exc)
        else:
            dependent_files = {Path(dep) for dep in files}
        os.environ["PATH"] = orig_path
        return dependent_files

    if LIEF_ENABLED:
        _get_dependent_files = _get_dependent_files_lief
    else:
        _get_dependent_files = _get_dependent_files_imagehlp

    def get_dependent_files(self, path: Union[str, Path]) -> Set[Path]:
        if isinstance(path, str):
            path = Path(path)
        with suppress(KeyError):
            return self.dependent_files[path]
        if not self.is_pe(path):
            return set()

        dependent_files: Set[Path] = self._get_dependent_files(path)
        self.dependent_files[path] = dependent_files
        return dependent_files

    def read_manifest(self, path: Union[str, Path]) -> str:
        if lief is None:
            raise RuntimeError("lief is not installed")
        if isinstance(path, str):
            path = Path(path)
        with path.open("rb", buffering=0) as raw:
            binary = lief.PE.parse(raw, path.name)
        try:
            resources_manager = binary.resources_manager
            manifest = resources_manager.manifest
        except lief.exception as exc:
            raise RuntimeError(exc) from None
        return manifest

    def write_manifest(self, path: Union[str, Path], manifest: str) -> None:
        if lief is None:
            raise RuntimeError("lief is not installed")
        if isinstance(path, str):
            path = Path(path)
        with path.open("rb", buffering=0) as raw:
            binary = lief.PE.parse(raw, path.name)
        try:
            resources_manager = binary.resources_manager
            resources_manager.manifest = manifest
            builder = lief.PE.Builder(binary)
            builder.build_resources(True)
            builder.build()
            with TemporaryPath("temp.exe") as tmp_path:
                builder.write(str(tmp_path))
                tmp_path.replace(path)
        except lief.exception as exc:
            raise RuntimeError(exc) from None


class ELFParser(Parser):
    """`ELFParser` is based on the logic around invoking `patchelf` and
    `ldd`."""

    def __init__(self, silent: int = 0) -> None:
        super().__init__(silent)
        self.linker_warnings: Dict[Path, Any] = {}
        _verify_patchelf()

    @staticmethod
    def is_elf(path: Union[str, Path]) -> bool:
        if isinstance(path, str):
            path = Path(path)
        if (
            path.suffix in NON_ELF_EXT
            or path.is_symlink()
            or not path.is_file()
        ):
            return False
        with open(path, "rb") as binary:
            four_bytes = binary.read(4)
        return bool(four_bytes == MAGIC_ELF)

    def get_dependent_files(self, path: Union[str, Path]) -> Set[Path]:
        if isinstance(path, str):
            path = Path(path)
        with suppress(KeyError):
            return self.dependent_files[path]

        dependent_files: Set[Path] = set()
        if not self.is_elf(path) or not os.access(path, os.X_OK):
            return dependent_files

        split_string = " => "
        dependent_file_index = 1
        args = ("ldd", path)
        process = run(
            args, check=False, stdout=PIPE, stderr=PIPE, encoding="utf-8"
        )
        for line in process.stdout.splitlines():
            parts = line.expandtabs().strip().split(split_string)
            if len(parts) != 2:
                continue
            dependent_file = parts[dependent_file_index].strip()
            if dependent_file == path.name:
                continue
            if dependent_file in ("not found", "(file not found)"):
                filename = parts[0]
                if filename not in self.linker_warnings:
                    self.linker_warnings[filename] = None
                    if self._silent < 3:
                        print(f"WARNING: cannot find '{filename}'")
                continue
            if dependent_file.startswith("("):
                continue
            pos = dependent_file.find(" (")
            if pos >= 0:
                dependent_file = dependent_file[:pos].strip()
            if dependent_file:
                dependent_files.add(Path(dependent_file))
        if process.returncode and self._silent < 3:
            print("WARNING:", *args, "returns:")
            print(process.stderr, end="")
        self.dependent_files[path] = dependent_files
        return dependent_files

    def get_rpath(self, filename: Union[str, Path]) -> str:
        args = ["patchelf", "--print-rpath", filename]
        with suppress(CalledProcessError):
            return check_output(args, encoding="utf-8").strip()
        return ""

    def replace_needed(
        self, filename: Union[str, Path], so_name: str, new_so_name: str
    ) -> None:
        self._set_write_mode(filename)
        args = ["patchelf", "--replace-needed", so_name, new_so_name, filename]
        check_call(args)

    def set_rpath(self, filename: Union[str, Path], rpath: str) -> None:
        self._set_write_mode(filename)
        args = ["patchelf", "--remove-rpath", filename]
        check_call(args)
        args = ["patchelf", "--force-rpath", "--set-rpath", rpath, filename]
        check_call(args)

    def set_soname(self, filename: Union[str, Path], new_so_name: str) -> None:
        self._set_write_mode(filename)
        args = ["patchelf", "--set-soname", new_so_name, filename]
        check_call(args)

    @staticmethod
    def _set_write_mode(filename: Union[str, Path]) -> None:
        if isinstance(filename, str):
            filename = Path(filename)
        mode = filename.stat().st_mode
        if mode & stat.S_IWUSR == 0:
            filename.chmod(mode | stat.S_IWUSR)


def _verify_patchelf() -> None:
    """This function looks for the ``patchelf`` external binary in the PATH,
    checks for the required version, and throws an exception if a proper
    version can't be found. Otherwise, silence is golden."""
    if not shutil.which("patchelf"):
        raise ValueError("Cannot find required utility `patchelf` in PATH")
    try:
        version = check_output(["patchelf", "--version"], encoding="utf-8")
    except CalledProcessError:
        raise ValueError("Could not call `patchelf` binary") from None

    mobj = re.match(r"patchelf\s+(\d+(.\d+)?)", version)
    if mobj and tuple(int(x) for x in mobj.group(1).split(".")) >= (0, 12):
        return
    raise ValueError(
        f"patchelf {version} found. cx-freeze requires patchelf >= 0.12."
    )



"""Create standalone executables from Python scripts, with the same performance
and is cross-platform."""
# pylint: disable=invalid-name

import setuptools  # isort:skip
import sys

from .command.build import Build as build
from .command.build_exe import BuildEXE as build_exe
from .command.install import Install as install
from .command.install_exe import InstallEXE as install_exe
from .dist import Distribution
from .exception import ConfigError
from .finder import Module, ModuleFinder
from .freezer import ConstantsModule, Executable, Freezer

__all__ = [
    "build",
    "build_exe",
    "install",
    "install_exe",
    "setup",
    "ConfigError",
    "ConstantsModule",
    "Executable",
    "Freezer",
    "Module",
    "ModuleFinder",
    "__version__",
]

if sys.platform == "win32":
    from .command.bdist_msi import BdistMSI as bdist_msi

    __all__.append(bdist_msi.__name__)
elif sys.platform == "darwin":
    from .command.bdist_mac import BdistDMG as bdist_dmg
    from .command.bdist_mac import BdistMac as bdist_mac

    __all__.extend([bdist_dmg.__name__, bdist_mac.__name__])
else:
    from .command.bdist_rpm import BdistRPM as bdist_rpm

    __all__.append(bdist_rpm.__name__)

__version__ = "6.11.1"


def _add_command_class(command_classes, name, cls):
    if name not in command_classes:
        command_classes[name] = cls


def setup(**attrs):  # pylint: disable=C0116
    attrs.setdefault("distclass", Distribution)
    command_classes = attrs.setdefault("cmdclass", {})
    if sys.platform == "win32":
        _add_command_class(command_classes, "bdist_msi", bdist_msi)
    elif sys.platform == "darwin":
        _add_command_class(command_classes, "bdist_dmg", bdist_dmg)
        _add_command_class(command_classes, "bdist_mac", bdist_mac)
    else:
        _add_command_class(command_classes, "bdist_rpm", bdist_rpm)
    _add_command_class(command_classes, "build", build)
    _add_command_class(command_classes, "build_exe", build_exe)
    _add_command_class(command_classes, "install", install)
    _add_command_class(command_classes, "install_exe", install_exe)
    setuptools.setup(**attrs)


setup.__doc__ = setuptools.setup.__doc__



"""Implements 'build' command.

Borrowed from distutils.command.build of Python 3.10 and merged with
build subclass of cx_Freeze 6.10.

"""

import os
import sys
from sysconfig import get_platform, get_python_version

from setuptools import Command
from setuptools.errors import OptionError

__all__ = ["Build"]


def show_compilers():
    """List available compilers."""
    try:
        distutils_ccompiler = __import__(
            "distutils.ccompiler", fromlist=["show_compilers"]
        )
        distutils_ccompiler.show_compilers()
    except (ImportError, AttributeError):
        print("The list of available compilers was not found.")


# pylint: disable=attribute-defined-outside-init,missing-function-docstring
class Build(Command):
    """Build everything needed to install."""

    description = "build everything needed to install"

    user_options = [
        ("build-base=", "b", "base directory for build library"),
        (
            "build-purelib=",
            None,
            "build directory for platform-neutral distributions",
        ),
        (
            "build-platlib=",
            None,
            "build directory for platform-specific distributions",
        ),
        (
            "build-lib=",
            None,
            "build directory for all distribution (defaults to either "
            + "build-purelib or build-platlib",
        ),
        ("build-scripts=", None, "build directory for scripts"),
        ("build-temp=", "t", "temporary build directory"),
        (
            "plat-name=",
            "p",
            "platform name to build for, if supported "
            f"(default: {get_platform()})",
        ),
        ("compiler=", "c", "specify the compiler type"),
        ("parallel=", "j", "number of parallel build jobs"),
        (
            "debug",
            "g",
            "compile extensions and libraries with debugging information",
        ),
        ("force", "f", "forcibly build everything (ignore file timestamps)"),
        (
            "executable=",
            "e",
            "specify final destination interpreter path (build.py)",
        ),
        ("build-exe=", None, "build directory for executables"),
    ]

    boolean_options = ["debug", "force"]

    help_options = [
        ("help-compiler", None, "list available compilers", show_compilers),
    ]

    def initialize_options(self):
        self.build_base = "build"
        # these are decided only after 'build_base' has its final value
        # (unless overridden by the user or client)
        self.build_purelib = None
        self.build_platlib = None
        self.build_lib = None
        self.build_temp = None
        self.build_scripts = None
        self.compiler = None
        self.plat_name = None
        self.debug = None
        self.force = 0
        self.executable = None
        self.parallel = None
        self.build_exe = None

    def finalize_options(self):
        platform = get_platform()
        python_version = get_python_version()
        if self.plat_name is None:
            self.plat_name = platform
        else:
            # plat-name only supported for windows (other platforms are
            # supported via ./configure flags, if at all).  Avoid misleading
            # other platforms.
            if os.name != "nt":
                raise OptionError(
                    "--plat-name only supported on Windows (try "
                    "using './configure --help' on your platform)"
                )

        plat_specifier = f".{self.plat_name}-{python_version}"

        # Make it so Python 2.x and Python 2.x with --with-pydebug don't
        # share the same build directories. Doing so confuses the build
        # process for C modules
        if hasattr(sys, "gettotalrefcount"):
            plat_specifier += "-pydebug"

        # 'build_purelib' and 'build_platlib' just default to 'lib' and
        # 'lib.<plat>' under the base build directory.  We only use one of
        # them for a given distribution, though --
        if self.build_purelib is None:
            self.build_purelib = os.path.join(self.build_base, "lib")
        if self.build_platlib is None:
            self.build_platlib = os.path.join(
                self.build_base, "lib" + plat_specifier
            )

        # 'build_lib' is the actual directory that we will use for this
        # particular module distribution -- if user didn't supply it, pick
        # one of 'build_purelib' or 'build_platlib'.
        if self.build_lib is None:
            if self.distribution.has_ext_modules():
                self.build_lib = self.build_platlib
            else:
                self.build_lib = self.build_purelib

        # 'build_temp' -- temporary directory for compiler turds,
        # "build/temp.<plat>"
        if self.build_temp is None:
            self.build_temp = os.path.join(
                self.build_base, "temp" + plat_specifier
            )
        if self.build_scripts is None:
            self.build_scripts = os.path.join(
                self.build_base, f"scripts-{python_version}"
            )

        if self.executable is None and sys.executable:
            self.executable = os.path.normpath(sys.executable)

        if isinstance(self.parallel, str):
            try:
                self.parallel = int(self.parallel)
            except ValueError:
                raise OptionError("parallel should be an integer") from None

        # 'build_exe' is the actual directory that we will use for this
        if self.build_exe is None:
            self.build_exe = os.path.join(
                self.build_base, "exe" + plat_specifier
            )

    def run(self):
        # Run all relevant sub-commands.  This will be some subset of:
        #  - build_py      - pure Python modules
        #  - build_clib    - standalone C libraries
        #  - build_ext     - Python extensions
        #  - build_scripts - (Python) scripts
        #  - build_exe     - executables
        for cmd_name in self.get_sub_commands():
            self.run_command(cmd_name)

    # -- Predicates for the sub-command list ---------------------------

    def has_pure_modules(self):
        return self.distribution.has_pure_modules()

    def has_c_libraries(self):
        return self.distribution.has_c_libraries()

    def has_ext_modules(self):
        return self.distribution.has_ext_modules()

    def has_scripts(self):
        return self.distribution.has_scripts()

    def has_executables(self):
        return self.distribution.has_executables()

    sub_commands = [
        ("build_py", has_pure_modules),
        ("build_clib", has_c_libraries),
        ("build_ext", has_ext_modules),
        ("build_scripts", has_scripts),
        ("build_exe", has_executables),
    ]



"""Implements the 'install_exe' command."""

import sys
from pathlib import Path
from typing import Optional

from setuptools import Command

__all__ = ["InstallEXE"]


# pylint: disable=attribute-defined-outside-init,missing-function-docstring
class InstallEXE(Command):
    """Install executables built from Python scripts."""

    description = "install executables built from Python scripts"
    user_options = [
        ("install-dir=", "d", "directory to install executables to"),
        ("build-dir=", "b", "build directory (where to install from)"),
        ("force", "f", "force installation (overwrite existing files)"),
        ("skip-build", None, "skip the build steps"),
    ]

    def initialize_options(self):
        self.install_dir: Optional[str] = None
        self.force = 0
        self.build_dir = None
        self.skip_build = None

    def finalize_options(self):
        self.set_undefined_options("build", ("build_exe", "build_dir"))
        self.set_undefined_options(
            "install",
            ("install_exe", "install_dir"),
            ("force", "force"),
            ("skip_build", "skip_build"),
        )

    def run(self):
        if not self.skip_build:
            self.run_command("build_exe")
        self.outfiles = self.copy_tree(self.build_dir, self.install_dir)
        if sys.platform != "win32":
            install_dir = Path(self.install_dir)
            base_dir = install_dir.parent.parent
            bin_dir: Path = base_dir / "bin"
            if not bin_dir.exists():
                bin_dir.mkdir(parents=True)
            source_dir = ".." / install_dir.relative_to(base_dir)
            for executable in self.distribution.executables:
                name = executable.target_name
                source = source_dir / name
                target = bin_dir / name
                if target.exists():
                    target.unlink()
                target.symlink_to(source)
                self.outfiles.append(target.as_posix())

    def get_inputs(self):
        return self.distribution.executables or []

    def get_outputs(self):
        return self.outfiles or []



"""Extends setuptools to build macOS dmg or app blundle."""

import os
import plistlib
import shutil
import subprocess
import warnings
from typing import List, Tuple

from setuptools import Command

from cx_Freeze.common import normalize_to_list
from cx_Freeze.darwintools import (
    DarwinFile,
    DarwinFileTracker,
    applyAdHocSignature,
    changeLoadReference,
)

__all__ = ["BdistDMG", "BdistMac"]


# pylint: disable=attribute-defined-outside-init,missing-function-docstring
class BdistDMG(Command):
    """Create a Mac DMG disk image containing the Mac application bundle."""

    description = (
        "create a Mac DMG disk image containing the Mac application bundle"
    )
    user_options = [
        ("volume-label=", None, "Volume label of the DMG disk image"),
        (
            "applications-shortcut=",
            None,
            "Boolean for whether to include "
            "shortcut to Applications in the DMG disk image",
        ),
        ("silent", "s", "suppress all output except warnings"),
    ]

    def initialize_options(self):
        self.volume_label = self.distribution.get_fullname()
        self.applications_shortcut = False
        self.silent = None

    def finalize_options(self):
        if self.silent is None:
            self.silent = False

    def build_dmg(self):
        # Remove DMG if it already exists
        if os.path.exists(self.dmg_name):
            os.unlink(self.dmg_name)

        # Make dist folder
        self.dist_dir = os.path.join(self.build_dir, "dist")
        if os.path.exists(self.dist_dir):
            shutil.rmtree(self.dist_dir)
        self.mkpath(self.dist_dir)

        # Copy App Bundle
        dest_dir = os.path.join(
            self.dist_dir, os.path.basename(self.bundle_dir)
        )
        self.copy_tree(self.bundle_dir, dest_dir)

        createargs = [
            "hdiutil",
            "create",
        ]
        if self.silent:
            createargs += ["-quiet"]
        createargs += [
            "-fs",
            "HFSX",
            "-format",
            "UDZO",
            self.dmg_name,
            "-imagekey",
            "zlib-level=9",
            "-srcfolder",
            self.dist_dir,
            "-volname",
            self.volume_label,
        ]

        if self.applications_shortcut:
            apps_folder_link = os.path.join(self.dist_dir, "Applications")
            os.symlink(
                "/Applications", apps_folder_link, target_is_directory=True
            )

        # Create the dmg
        if subprocess.call(createargs) != 0:
            raise OSError("creation of the dmg failed")

    def run(self):
        # Create the application bundle
        self.run_command("bdist_mac")

        # Find the location of the application bundle and the build dir
        self.bundle_dir = self.get_finalized_command("bdist_mac").bundle_dir
        self.build_dir = self.get_finalized_command("build").build_base

        # Set the file name of the DMG to be built
        self.dmg_name = os.path.join(
            self.build_dir, self.volume_label + ".dmg"
        )

        self.execute(self.build_dmg, ())


class BdistMac(Command):
    """Create a Mac application bundle."""

    description = "create a Mac application bundle"

    plist_items: List[Tuple[str, str]]
    include_frameworks: List[str]
    include_resources: List[str]

    user_options = [
        ("iconfile=", None, "Path to an icns icon file for the application."),
        (
            "qt-menu-nib=",
            None,
            "Location of qt_menu.nib folder for Qt "
            "applications. Will be auto-detected by default.",
        ),
        (
            "bundle-name=",
            None,
            "File name for the bundle application "
            "without the .app extension.",
        ),
        (
            "plist-items=",
            None,
            "A list of key-value pairs (type: List[Tuple[str, str]]) to "
            "be added to the app bundle Info.plist file.",
        ),
        (
            "custom-info-plist=",
            None,
            "File to be used as the Info.plist in "
            "the app bundle. A basic one will be generated by default.",
        ),
        (
            "include-frameworks=",
            None,
            "A comma separated list of Framework "
            "directories to include in the app bundle.",
        ),
        (
            "include-resources=",
            None,
            "A list of tuples of additional "
            "files to include in the app bundle's resources directory, with "
            "the first element being the source, and second the destination "
            "file or directory name.",
        ),
        (
            "codesign-identity=",
            None,
            "The identity of the key to be used to sign the app bundle.",
        ),
        (
            "codesign-entitlements=",
            None,
            "The path to an entitlements file "
            "to use for your application's code signature.",
        ),
        (
            "codesign-deep=",
            None,
            "Boolean for whether to codesign using the --deep option.",
        ),
        (
            "codesign-resource-rules",
            None,
            "Plist file to be passed to "
            "codesign's --resource-rules option.",
        ),
        (
            "absolute-reference-path=",
            None,
            "Path to use for all referenced "
            "libraries instead of @executable_path.",
        ),
        (
            "rpath-lib-folder",
            None,
            "DEPRECATED.  Will be removed in next version.",
        ),
    ]

    def initialize_options(self):
        self.list_options = [
            "plist_items",
            "include_frameworks",
            "include_resources",
        ]
        for option in self.list_options:
            setattr(self, option, [])

        self.iconfile = None
        self.qt_menu_nib = False
        self.bundle_name = self.distribution.get_fullname()
        self.custom_info_plist = None
        self.codesign_identity = None
        self.codesign_entitlements = None
        self.codesign_deep = None
        self.codesign_resource_rules = None
        self.absolute_reference_path = None
        self.rpath_lib_folder = None

    def finalize_options(self):
        # Make sure all options of multiple values are lists
        for option in self.list_options:
            setattr(self, option, normalize_to_list(getattr(self, option)))
        for item in self.plist_items:
            if not isinstance(item, tuple) or len(item) != 2:
                raise Exception(
                    "Error, plist_items must be a list of key, value pairs "
                    "(List[Tuple[str, str]]) (bad list item)."
                )
        if self.rpath_lib_folder is not None:
            warnings.warn(
                "rpath-lib-folder is obsolete and will be removed in the "
                "next major version"
            )

    def create_plist(self):
        """Create the Contents/Info.plist file"""
        # Use custom plist if supplied, otherwise create a simple default.
        if self.custom_info_plist:
            with open(self.custom_info_plist, "rb") as file:
                contents = plistlib.load(file)
        else:
            contents = {
                "CFBundleIconFile": "icon.icns",
                "CFBundleDevelopmentRegion": "English",
                "CFBundleIdentifier": self.bundle_name,
                # Specify that bundle is an application bundle
                "CFBundlePackageType": "APPL",
                # Cause application to run in high-resolution mode by default
                # (Without this, applications run from application bundle may
                # be pixelated)
                "NSHighResolutionCapable": "True",
            }

        # Ensure CFBundleExecutable is set correctly
        contents["CFBundleExecutable"] = self.bundle_executable

        # add custom items to the plist file
        for key, value in self.plist_items:
            contents[key] = value

        with open(os.path.join(self.contents_dir, "Info.plist"), "wb") as file:
            plistlib.dump(contents, file)

    def set_absolute_reference_paths(self, path=None):
        """
        For all files in Contents/MacOS, set their linked library paths to be
        absolute paths using the given path instead of @executable_path.
        """
        if not path:
            path = self.absolute_reference_path

        files = os.listdir(self.bin_dir)

        for filename in files:
            filename = os.path.join(self.bin_dir, filename)

            # Skip some file types
            if filename[-1:] in ("txt", "zip") or os.path.isdir(filename):
                continue

            out = subprocess.check_output(
                ("otool", "-L", filename), encoding="utf-8"
            )
            for line in out.splitlines()[1:]:
                lib = line.lstrip("\t").split(" (compat")[0]

                if lib.startswith("@executable_path"):
                    replacement = lib.replace("@executable_path", path)

                    path, name = os.path.split(replacement)

                    # see if we provide the referenced file;
                    # if so, change the reference
                    if name in files:
                        subprocess.call(
                            (
                                "install_name_tool",
                                "-change",
                                lib,
                                replacement,
                                filename,
                            )
                        )
            applyAdHocSignature(filename)

    def set_relative_reference_paths(self, build_dir: str, bin_dir: str):
        """
        Make all the references from included Mach-O files to other included
        Mach-O files relative.
        """
        darwin_file: DarwinFile

        for darwin_file in self.darwin_tracker:
            # get the relative path to darwin_file in build directory
            relative_copy_dest = os.path.relpath(
                darwin_file.getBuildPath(), build_dir
            )
            # figure out directory where it will go in binary directory for
            # .app bundle, this would be the Content/MacOS subdirectory in
            # bundle.  This is the file that needs to have its dynamic load
            # references updated.
            file_path_in_bin_dir = os.path.join(bin_dir, relative_copy_dest)

            # for each file that this darwin_file references, update the
            # reference as necessary; if the file is copied into the binary
            # package, change the reference to be relative to @executable_path
            # (so an .app bundle will work wherever it is moved)
            for reference in darwin_file.getMachOReferenceList():
                if not reference.is_copied:
                    # referenced file not copied -- assume this is a system
                    # file that will also be present on the user's machine,
                    # and do not change reference
                    continue
                # this is the reference in the machO file that needs to be
                # updated
                raw_path = reference.raw_path
                ref_target_file: DarwinFile = reference.target_file
                # this is where file copied in build dir
                abs_build_dest = ref_target_file.getBuildPath()
                rel_build_dest = os.path.relpath(abs_build_dest, build_dir)
                exe_path = f"@executable_path/{rel_build_dest}"
                changeLoadReference(
                    file_path_in_bin_dir,
                    oldReference=raw_path,
                    newReference=exe_path,
                    VERBOSE=False,
                )

            applyAdHocSignature(file_path_in_bin_dir)

    def find_qt_menu_nib(self):
        """
        Returns a location of a qt_menu.nib folder, or None if this is not
        a Qt application.
        """
        if self.qt_menu_nib:
            return self.qt_menu_nib
        if any(n.startswith("PyQt4.QtCore") for n in os.listdir(self.bin_dir)):
            name = "PyQt4"
        elif any(
            n.startswith("PySide.QtCore") for n in os.listdir(self.bin_dir)
        ):
            name = "PySide"
        else:
            return None

        qtcore = __import__(name, fromlist=["QtCore"]).QtCore
        libpath = str(
            qtcore.QLibraryInfo.location(qtcore.QLibraryInfo.LibrariesPath)
        )
        for subpath in [
            "QtGui.framework/Resources/qt_menu.nib",
            "Resources/qt_menu.nib",
        ]:
            path = os.path.join(libpath, subpath)
            if os.path.exists(path):
                return path

        # Last resort: fixed paths (macports)
        for path in [
            "/opt/local/Library/Frameworks/QtGui.framework/Versions/"
            "4/Resources/qt_menu.nib"
        ]:
            if os.path.exists(path):
                return path

        print("Could not find qt_menu.nib")
        raise OSError("Could not find qt_menu.nib")

    def prepare_qt_app(self):
        """
        Add resource files for a Qt application. Should do nothing if the
        application does not use QtCore.
        """
        nib_locn = self.find_qt_menu_nib()
        if nib_locn is None:
            return

        # Copy qt_menu.nib
        self.copy_tree(
            nib_locn, os.path.join(self.resources_dir, "qt_menu.nib")
        )

        # qt.conf needs to exist, but needn't have any content
        with open(os.path.join(self.resources_dir, "qt.conf"), "wb"):
            pass

    def run(self):
        self.run_command("build")
        build = self.get_finalized_command("build")
        freezer: "freezer.Freezer" = self.get_finalized_command(
            "build_exe"
        ).freezer

        # Define the paths within the application bundle
        self.bundle_dir = os.path.join(
            build.build_base, self.bundle_name + ".app"
        )
        self.contents_dir = os.path.join(self.bundle_dir, "Contents")
        self.resources_dir = os.path.join(self.contents_dir, "Resources")
        self.bin_dir = os.path.join(self.contents_dir, "MacOS")
        self.frameworks_dir = os.path.join(self.contents_dir, "Frameworks")

        # Find the executable name
        executable = self.distribution.executables[0].target_name
        _, self.bundle_executable = os.path.split(executable)

        # Build the app directory structure
        self.mkpath(self.resources_dir)
        self.mkpath(self.bin_dir)
        self.mkpath(self.frameworks_dir)

        self.copy_tree(build.build_exe, self.bin_dir)

        # Copy the icon
        if self.iconfile:
            self.copy_file(
                self.iconfile, os.path.join(self.resources_dir, "icon.icns")
            )

        # Copy in Frameworks
        for framework in self.include_frameworks:
            self.copy_tree(
                framework,
                os.path.join(self.frameworks_dir, os.path.basename(framework)),
            )

        # Copy in Resources
        for resource, destination in self.include_resources:
            if os.path.isdir(resource):
                self.copy_tree(
                    resource, os.path.join(self.resources_dir, destination)
                )
            else:
                parent_dirs = os.path.dirname(
                    os.path.join(self.resources_dir, destination)
                )
                os.makedirs(parent_dirs, exist_ok=True)
                self.copy_file(
                    resource, os.path.join(self.resources_dir, destination)
                )

        # Create the Info.plist file
        self.execute(self.create_plist, ())

        # Make all references to libraries relative
        self.darwin_tracker: DarwinFileTracker = freezer.darwin_tracker
        self.execute(
            self.set_relative_reference_paths,
            (os.path.abspath(build.build_exe), os.path.abspath(self.bin_dir)),
        )

        # Make library references absolute if enabled
        if self.absolute_reference_path:
            self.execute(self.set_absolute_reference_paths, ())

        # For a Qt application, run some tweaks
        self.execute(self.prepare_qt_app, ())

        # Sign the app bundle if a key is specified
        if self.codesign_identity:
            signargs = ["codesign", "-s", self.codesign_identity]

            if self.codesign_entitlements:
                signargs.append("--entitlements")
                signargs.append(self.codesign_entitlements)

            if self.codesign_deep:
                signargs.insert(1, "--deep")

            if self.codesign_resource_rules:
                signargs.insert(
                    1, "--resource-rules=" + self.codesign_resource_rules
                )

            signargs.append(self.bundle_dir)

            if subprocess.call(signargs) != 0:
                raise OSError("Code signing of app bundle failed")



"""The classes and functions with which cx_Freeze extends setuptools."""

from setuptools import Distribution as _Distribution

__all__ = ["Distribution", "DistributionMetadata"]


class DistributionMetadata:
    """Dummy class to hold the distribution meta-data: name, version,
    author, and so forth."""


class Distribution(_Distribution):
    """Distribution with support for executables."""

    def __init__(self, attrs):
        self.executables = []
        super().__init__(attrs)
        self.metadata = self._patch_metadata()

    def _patch_metadata(self) -> DistributionMetadata:
        old_metadata = self.metadata
        new_metadata = DistributionMetadata()
        for var in dir(old_metadata):
            if not hasattr(new_metadata, var):
                setattr(new_metadata, var, getattr(old_metadata, var))
        self.distutils_metadata = old_metadata
        return new_metadata

    def has_executables(self):
        """Predicate for build_exe command."""
        return self.executables and len(self.executables) > 0



"""Common utility functions shared between cx_Freeze modules."""

import contextlib
import shutil
import types
import warnings
from pathlib import Path, PurePath
from tempfile import TemporaryDirectory
from typing import List, Optional, Tuple, Union

from .exception import ConfigError

IncludesList = List[
    Union[str, Path, Tuple[Union[str, Path], Optional[Union[str, Path]]]]
]
InternalIncludesList = List[Tuple[Path, PurePath]]


class FilePath(Path):
    """Subclass of concrete Path to be used in TemporaryPath."""

    _flavour = type(Path())._flavour  # pylint: disable=E1101,W0212

    def replace(self, target):
        """Rename this path to the target path, overwriting if that path
        exists. Extended to support move between file systems."""
        with contextlib.suppress(OSError):
            return super().replace(target)
        shutil.copyfile(self, target)
        with contextlib.suppress(FileNotFoundError):
            self.unlink()
        return self.__class__(target)


class TemporaryPath(TemporaryDirectory):
    """Create and return a Path-like temporary directory."""

    def __init__(
        self, filename=None, suffix=None, prefix=None, dir=None
    ):  # pylint: disable=redefined-builtin
        super().__init__(suffix, prefix or "cxfreeze-", dir)
        if filename:
            if Path(filename).parent.name:
                raise ValueError("filename cannot contain directory")
            self.path = FilePath(self.name, filename)
        else:
            self.path = FilePath(self.name)

    def __enter__(self):
        return self.path


def get_resource_file_path(
    dirname: Union[str, Path], name: Union[str, Path], ext: str
) -> Optional[Path]:
    """
    Return the path to a resource file shipped with cx_Freeze.

    This is used to find our base executables and initscripts when they are
    just specified by name.
    """
    pname = Path(name)
    if pname.is_absolute():
        return pname
    pname = Path(__file__).resolve().parent / dirname / pname.with_suffix(ext)
    if pname.exists():
        return pname
    return None


def normalize_to_list(
    value: Optional[Union[str, List[str], Tuple[str, ...]]]
) -> List[str]:
    """
    Takes the different formats of options containing multiple values and
    returns the value as a list object.
    """
    if value is None:
        normalized_value = []
    elif isinstance(value, str):
        normalized_value = value.split(",")
    else:
        normalized_value = list(value)

    return normalized_value


def process_path_specs(specs: Optional[IncludesList]) -> InternalIncludesList:
    """
    Prepare paths specified as config.

    The input is a list of either strings, or 2-tuples (source, target).
    Where single strings are supplied, the basenames are used as targets.
    Where targets are given explicitly, they must not be absolute paths.

    Returns a list of 2-tuples, or throws ConfigError if something is wrong
    in the input.
    """
    if specs is None:
        specs = []
    processed_specs: InternalIncludesList = []
    for spec in specs:
        if not isinstance(spec, (list, tuple)):
            source = spec
            target = None
        elif len(spec) != 2:
            error = "path spec must be a list or tuple of length two"
            raise ConfigError(error)
        else:
            source, target = spec
        source = Path(source)
        if not source.exists():
            raise ConfigError(f"cannot find file/directory named {source!s}")
        target = PurePath(target or source.name)
        if target.is_absolute():
            error = f"target path named {target!s} cannot be absolute"
            raise ConfigError(error)
        processed_specs.append((source, target))
    return processed_specs


def code_object_replace(code: types.CodeType, **kwargs) -> types.CodeType:
    """
    Return a copy of the code object with new values for the specified fields.
    """
    try:
        kwargs["co_consts"] = tuple(kwargs["co_consts"])
    except ValueError:
        pass
    # Python 3.8+
    if hasattr(code, "replace"):
        return code.replace(**kwargs)
    params = [
        kwargs.get("co_argcount", code.co_argcount),
        kwargs.get("co_kwonlyargcount", code.co_kwonlyargcount),
        kwargs.get("co_nlocals", code.co_nlocals),
        kwargs.get("co_stacksize", code.co_stacksize),
        kwargs.get("co_flags", code.co_flags),
        kwargs.get("co_code", code.co_code),
        kwargs.get("co_consts", code.co_consts),
        kwargs.get("co_names", code.co_names),
        kwargs.get("co_varnames", code.co_varnames),
        kwargs.get("co_filename", code.co_filename),
        kwargs.get("co_name", code.co_name),
        kwargs.get("co_firstlineno", code.co_firstlineno),
        kwargs.get("co_lnotab", code.co_lnotab),
        kwargs.get("co_freevars", code.co_freevars),
        kwargs.get("co_cellvars", code.co_cellvars),
    ]
    return types.CodeType(*params)


def validate_args(arg, snake_value, camel_value):
    """
    Validate arguments from two exclusive sources.
    This is a temporary function to be used while transitioning from using
    camelCase parameters to snake_case.
    """
    if isinstance(snake_value, str):
        if isinstance(camel_value, str):
            raise ConfigError(
                f"May not pass {arg!r} as snake_case and camelCase"
            )
    elif isinstance(camel_value, str):
        warnings.warn(
            "camelCase values is obsolete and will be removed in the "
            f"next major version -> use the new name {arg!r}"
        )
    return snake_value or camel_value



import os
import subprocess


class SetupWriter:
    bases = {"C": "Console", "G": "Win32GUI", "S": "Win32Service"}

    @property
    def base(self):
        return self.bases[self.base_code]

    @property
    def default_executable_name(self):
        return os.path.splitext(self.script)[0]

    def __init__(self):
        self.name = self.description = self.script = ""
        self.executable_name = self.default_executable_name
        self.setup_file_name = "setup.py"
        self.version = "1.0"
        self.base_code = "C"

    def get_boolean_value(self, label, default=False):
        default_response = "y" if default else "n"
        while True:
            response = self.get_value(
                label, default_response, separator="? "
            ).lower()
            if response in ("y", "n", "yes", "no"):
                break
        return response in ("y", "yes")

    def get_value(self, label, default="", separator=": "):
        if default:
            label += " [%s]" % default
        return input(label + separator).strip() or default

    def populate_from_command_line(self):
        self.name = self.get_value("Project name", self.name)
        self.version = self.get_value("Version", self.version)
        self.description = self.get_value("Description", self.description)
        self.script = self.get_value(
            "Python file to make executable from", self.script
        )
        self.executable_name = self.get_value(
            "Executable file name", self.default_executable_name
        )
        bases_prompt = "(C)onsole application, (G)UI application, or (S)ervice"
        while True:
            self.base_code = self.get_value(bases_prompt, "C")
            if self.base_code in self.bases:
                break
        while True:
            self.setup_file_name = self.get_value(
                "Save setup script to", self.setup_file_name
            )
            if not os.path.exists(self.setup_file_name):
                break
            if self.get_boolean_value("Overwrite %s" % self.setup_file_name):
                break

    def write(self):
        with open(self.setup_file_name, "w", encoding="UTF-8") as output:
            w = lambda s: output.write(s + "\n")  # noqa

            w("from cx_Freeze import setup, Executable")
            w("")

            w("# Dependencies are automatically detected, but it might need")
            w("# fine tuning.")
            w("build_options = {'packages': [], 'excludes': []}")
            w("")

            if self.base.startswith("Win32"):
                w("import sys")
                w("base = %r if sys.platform=='win32' else None" % self.base)
            else:
                w("base = %r" % self.base)
            w("")

            w("executables = [")
            if self.executable_name != self.default_executable_name:
                w(
                    "    Executable(%r, base=base, target_name = %r)"
                    % (self.script, self.executable_name)
                )
            else:
                w("    Executable(%r, base=base)" % self.script)
            w("]")
            w("")

            w(
                (
                    "setup(name=%r,\n"
                    "      version = %r,\n"
                    "      description = %r,\n"
                    "      options = {'build_exe': build_options},\n"
                    "      executables = executables)"
                )
                % (self.name, self.version, self.description)
            )


def main():
    writer = SetupWriter()
    writer.populate_from_command_line()
    writer.write()
    print("")
    print("Setup script written to %s; run it as:" % writer.setup_file_name)
    print("    python %s build" % writer.setup_file_name)
    if writer.get_boolean_value("Run this now"):
        subprocess.call(["python", writer.setup_file_name, "build"])



"""Internal exception classes."""

__all__ = ["ConfigError"]


class ConfigError(Exception):
    """Raised when an error is detected in the configuration.
    The associated value is a string indicating what precisely went wrong."""

    def __init__(self, msg: str):
        self.what: str = msg
        super().__init__()

    def __str__(self) -> str:
        return self.what


class DarwinException(Exception):
    """Raied when an error is detected in the module darwintools."""



import os
import platform
import shutil
import stat
import subprocess
import tempfile
from pathlib import Path
from typing import Dict, Iterable, List, Optional, Set, Union

from .exception import DarwinException

# In a MachO file, need to deal specially with links that use @executable_path,
# @loader_path, @rpath
#
# @executable_path - where ultimate calling executable is
# @loader_path - directory of current object
# @rpath - list of paths to check
# (earlier rpaths have higher priority, i believe)
#
# Resolving these variables (particularly @rpath) requires tracing through the
# sequence linked MachO files leading the the current file, to determine which
# directories are included in the current rpath.


def _isMachOFile(path: Path) -> bool:
    """Determines whether the file is a Mach-O file."""
    if not path.is_file():
        return False
    if b"Mach-O" in subprocess.check_output(("file", path)):
        return True
    return False


class MachOReference:
    """Represents a linking reference from MachO file to another file."""

    def __init__(
        self,
        source_file: "DarwinFile",
        raw_path: str,
        resolved_path: Optional[Path],
    ):
        """
        :param source_file: DarwinFile object for file in which the reference
        was found
        :param raw_path: The load path that appears in the file
        (may include @rpath, etc.)
        :param resolved_path: The path resolved to an explicit path to a file
        on system. Or None, if the path could not be resolved at the time the
        DarwinFile was processed.
        """
        self.source_file: "DarwinFile" = source_file
        self.raw_path: str = raw_path
        self.resolved_path: Optional[Path] = resolved_path

        # True if the referenced file is copied into the frozen package
        # (i.e., not a non-copied system file)
        self.is_copied = False
        # reference to target DarwinFile (but only if file is copied into app)
        self.target_file: Optional["DarwinFile"] = None

    def isResolved(self) -> bool:
        return self.resolved_path is not None

    def setTargetFile(self, darwin_file: "DarwinFile"):
        self.target_file = darwin_file
        self.is_copied = True


class DarwinFile:
    """A DarwinFile object represents a file that will be copied into the
    application, and record where it was ultimately moved to in the application
    bundle. Mostly used to provide special handling for copied files that are
    Mach-O files."""

    def __init__(
        self,
        path: Union[str, Path],
        referencing_file: Optional["DarwinFile"] = None,
        strict: bool = False,
    ):
        """
        :param path: The original path of the DarwinFile
        (before copying into app)
        :param referencing_file: DarwinFile object representing the referencing
        source file
        :param strict: Do not make guesses about rpath resolution. If the
        load does not resolve, throw an Exception.
        """
        self.path = Path(path).resolve()
        self.referencing_file: Optional["DarwinFile"] = None
        self.strict = strict

        # path to file in build directory (set as part of freeze process)
        self._build_path: Optional[Path] = None

        # commands in a Mach-O file
        self.commands: List[MachOCommand] = []
        self.loadCommands: List[MachOLoadCommand] = []
        self.rpathCommands: List[MachORPathCommand] = []

        # note: if file gets referenced twice (or more), it will only be the
        # first reference that gets recorded.
        # mapping of raw load paths to absolute resolved paths
        # (or None, if no resolution was determined)
        self.libraryPathResolution: Dict[str, Optional[Path]] = {}
        # the is of entries in the rpath in effect for this file.
        self._rpath: Optional[List[Path]] = None

        # dictionary of MachOReference objects, by their paths.
        # Path used is the resolved path, if available, and otherwise the
        # unresolved load path.
        self.machOReferenceForTargetPath: Dict[Path, MachOReference] = {}

        if not _isMachOFile(self.path):
            self.isMachO = False
            return

        # if this is a MachO file, extract linking information from it
        self.isMachO = True
        self.commands = MachOCommand._getMachOCommands(self.path)
        self.loadCommands = [
            c for c in self.commands if isinstance(c, MachOLoadCommand)
        ]
        self.rpathCommands = [
            c for c in self.commands if isinstance(c, MachORPathCommand)
        ]
        self.referencing_file = referencing_file

        self.getRPath()
        self.resolveLibraryPaths()

        # Create MachOReference objects for all the binaries referenced form
        # this file.
        for raw_path, resolved_path in self.libraryPathResolution.items():
            # the path to use for storing in dictionary
            if resolved_path is None:
                dict_path = Path(raw_path)
            else:
                dict_path = resolved_path
            if dict_path in self.machOReferenceForTargetPath:
                raise DarwinException(
                    "Multiple dynamic libraries resolved to the same file."
                )
            reference = MachOReference(
                source_file=self,
                raw_path=raw_path,
                resolved_path=resolved_path,
            )
            self.machOReferenceForTargetPath[dict_path] = reference

    def __str__(self):
        parts = []
        # parts.append("RPath Commands: {}".format(self.rpathCommands))
        # parts.append("Load commands: {}".format(self.loadCommands))
        parts.append(f"Mach-O File: {self.path}")
        parts.append("Resolved rpath:")
        for rpath in self.getRPath():
            parts.append(f"   {rpath}")
        parts.append("Loaded libraries:")
        for rpath in self.libraryPathResolution:
            parts.append(f"   {rpath} -> {self.libraryPathResolution[rpath]}")
        return "\n".join(parts)

    def fileReferenceDepth(self) -> int:
        """Returns how deep this Mach-O file is in the dynamic load order."""
        if self.referencing_file is not None:
            return self.referencing_file.fileReferenceDepth() + 1
        return 0

    def printFileInformation(self):
        """Prints information about the Mach-O file."""
        print(f"[{self.fileReferenceDepth()}] File: {self.path}")
        print("  Commands:")
        if len(self.commands) > 0:
            for cmd in self.commands:
                print(f"    {cmd}")
        else:
            print("    [None]")

        # This can be included for even more detail on the problem file.
        # print("  Load commands:")
        # if len(self.loadCommands) > 0:
        #     for cmd in self.loadCommands: print(f'    {cmd}')
        # else: print("    [None]")

        print("  RPath commands:")
        if len(self.rpathCommands) > 0:
            for rpc in self.rpathCommands:
                print(f"    {rpc}")
        else:
            print("    [None]")
        print("  Calculated RPath:")
        rpath = self.getRPath()
        if len(rpath) > 0:
            for path in rpath:
                print(f"    {path}")
        else:
            print("    [None]")
        if self.referencing_file is not None:
            print("Referenced from:")
            self.referencing_file.printFileInformation()

    def setBuildPath(self, path: Path):
        self._build_path = path

    def getBuildPath(self) -> Optional[Path]:
        return self._build_path

    @staticmethod
    def isExecutablePath(path: str) -> bool:
        return path.startswith("@executable_path")

    @staticmethod
    def isLoaderPath(path: str) -> bool:
        return path.startswith("@loader_path")

    @staticmethod
    def isRPath(path: str) -> bool:
        return path.startswith("@rpath")

    def resolveLoader(self, path: str) -> Optional[Path]:
        """Resolve a path that includes @loader_path. @loader_path represents
        the directory in which the DarwinFile is located."""
        if self.isLoaderPath(path):
            if path == "@loader_path":
                return self.path.parent
            return self.path.parent / path.replace("@loader_path/", "", 1)
        raise DarwinException(f"resolveLoader() called on bad path: {path}")

    def resolveExecutable(self, path: str) -> Path:
        """@executable_path should resolve to the directory where the original
        executable was located. By default, we set that to the directory of
        the library, so it would resolve in the same was as if linked from an
        executable in the same directory."""
        # consider making this resolve to the directory of the python
        # interpreter? Apparently not a big issue in practice, since the
        # code has been like this forever.
        if self.isExecutablePath(path):
            if path == "@executable_path":
                return self.path.parent
            return self.path.parent / path.replace("@executable_path/", "", 1)
        raise DarwinException(
            f"resolveExecutable() called on bad path: {path}"
        )

    def resolveRPath(self, path: str) -> Optional[Path]:
        for rpath in self.getRPath():
            if path == "@rpath":
                test_path = rpath
            else:
                test_path = rpath / path.replace("@rpath/", "", 1)
            if _isMachOFile(test_path):
                return test_path
        if not self.strict:
            # If not strictly enforcing rpath, return None here, and leave any
            # error to .finalizeReferences() instead.
            return None
        print(f"\nERROR: Problem resolving RPath [{path}] in file:")
        self.printFileInformation()
        raise DarwinException(f"resolveRPath() failed to resolve path: {path}")

    def getRPath(self) -> List[Path]:
        """Returns the rpath in effect for this file. Determined by rpath
        commands in this file and (recursively) the chain of files that
        referenced this file."""
        if self._rpath is not None:
            return self._rpath
        raw_paths = [c.rpath for c in self.rpathCommands]
        rpath = []
        for raw_path in raw_paths:
            test_rp = Path(raw_path)
            if test_rp.is_absolute():
                rpath.append(test_rp)
            elif self.isLoaderPath(raw_path):
                rpath.append(self.resolveLoader(raw_path).resolve())
            elif self.isExecutablePath(raw_path):
                rpath.append(self.resolveExecutable(raw_path).resolve())
        rpath = [raw_path for raw_path in rpath if raw_path.exists()]

        if self.referencing_file is not None:
            rpath = self.referencing_file.getRPath() + rpath
        self._rpath = rpath
        return rpath

    def resolvePath(self, path: str) -> Optional[Path]:
        """Resolves any @executable_path, @loader_path, and @rpath references
        in a path."""
        if self.isLoaderPath(path):  # replace @loader_path
            return self.resolveLoader(path)
        if self.isExecutablePath(path):  # replace @executable_path
            return self.resolveExecutable(path)
        if self.isRPath(path):  # replace @rpath
            return self.resolveRPath(path)
        test_path = Path(path)
        if test_path.is_absolute():  # just use the path, if it is absolute
            return test_path
        test_path = self.path.parent / path
        if _isMachOFile(test_path):
            return test_path.resolve()
        raise DarwinException(f"Could not resolve path: {path}")

    def resolveLibraryPaths(self):
        for cmd in self.loadCommands:
            raw_path = cmd.load_path
            resolved_path = self.resolvePath(raw_path)
            self.libraryPathResolution[raw_path] = resolved_path

    def getDependentFilePaths(self) -> Set[Path]:
        """Returns a list the available resolved paths to dependencies."""
        dependents: Set[Path] = set()
        for ref in self.machOReferenceForTargetPath.values():
            # skip load references that could not be resolved
            if ref.isResolved():
                dependents.add(ref.resolved_path)
        return dependents

    def getMachOReferenceList(self) -> List[MachOReference]:
        return list(self.machOReferenceForTargetPath.values())

    def getMachOReferenceForPath(self, path: Path) -> MachOReference:
        """Returns the reference pointing to the specified path, baed on paths
        stored in self.machOReferenceTargetPath. Raises Exception if not
        available."""
        try:
            return self.machOReferenceForTargetPath[path]
        except KeyError:
            raise DarwinException(
                f"Path {path} is not a path referenced from DarwinFile"
            ) from None


class MachOCommand:
    """Represents a load command in a MachO file."""

    def __init__(self, lines: List[str]):
        self.lines = lines

    def displayString(self) -> str:
        parts: List[str] = []
        if len(self.lines) > 0:
            parts.append(self.lines[0].strip())
        if len(self.lines) > 1:
            parts.append(self.lines[1].strip())
        return " / ".join(parts)

    def __repr__(self):
        return f"<MachOCommand ({self.displayString()})>"

    @staticmethod
    def _getMachOCommands(path: Path) -> List["MachOCommand"]:
        """Returns a list of load commands in the specified file, using
        otool."""
        shell_command = ("otool", "-l", path)
        commands: List[MachOCommand] = []
        current_command_lines = None

        # split the output into separate load commands
        out = subprocess.check_output(shell_command, encoding="utf-8")
        for line in out.splitlines():
            line = line.strip()
            if line[:12] == "Load command":
                if current_command_lines is not None:
                    commands.append(
                        MachOCommand.parseLines(current_command_lines)
                    )
                current_command_lines = []
            if current_command_lines is not None:
                current_command_lines.append(line)
        if current_command_lines is not None:
            commands.append(MachOCommand.parseLines(current_command_lines))
        return commands

    @staticmethod
    def parseLines(lines: List[str]) -> "MachOCommand":
        if len(lines) < 2:
            return MachOCommand(lines)
        parts = lines[1].split(" ")
        if parts[0] != "cmd":
            return MachOCommand(lines)
        if parts[1] == "LC_LOAD_DYLIB":
            return MachOLoadCommand(lines)
        if parts[1] == "LC_RPATH":
            return MachORPathCommand(lines)
        return MachOCommand(lines)


class MachOLoadCommand(MachOCommand):
    def __init__(self, lines: List[str]):
        super().__init__(lines)
        self.load_path = None
        if len(self.lines) < 4:
            return
        pathline = self.lines[3]
        pathline = pathline.strip()
        if not pathline.startswith("name "):
            return
        pathline = pathline[4:].strip()
        pathline = pathline.split("(offset")[0].strip()
        self.load_path = pathline

    def getPath(self):
        return self.load_path

    def __repr__(self):
        return f"<LoadCommand path={self.load_path!r}>"


class MachORPathCommand(MachOCommand):
    def __init__(self, lines: List[str]):
        super().__init__(lines)
        self.rpath = None
        if len(self.lines) < 4:
            return
        pathline = self.lines[3]
        pathline = pathline.strip()
        if not pathline.startswith("path "):
            return
        pathline = pathline[4:].strip()
        pathline = pathline.split("(offset")[0].strip()
        self.rpath = pathline

    def __repr__(self):
        return f"<RPath path={self.rpath!r}>"


def _printFile(
    darwinFile: DarwinFile,
    seenFiles: Set[DarwinFile],
    level: int,
    noRecurse=False,
):
    """
    Utility function to prints details about a DarwinFile and (optionally)
    recursively any other DarwinFiles that it references.
    """
    print("{}{}".format(level * "|  ", str(darwinFile.path)), end="")
    print(" (already seen)" if noRecurse else "")
    if noRecurse:
        return
    for ref in darwinFile.machOReferenceForTargetPath.values():
        if not ref.is_copied:
            continue
        file = ref.target_file
        _printFile(
            file,
            seenFiles=seenFiles,
            level=level + 1,
            noRecurse=(file in seenFiles),
        )
        seenFiles.add(file)
    return


def printMachOFiles(fileList: List[DarwinFile]):
    seenFiles = set()
    for file in fileList:
        if file not in seenFiles:
            seenFiles.add(file)
            _printFile(file, seenFiles=seenFiles, level=0)


def changeLoadReference(
    fileName: str, oldReference: str, newReference: str, VERBOSE: bool = True
):
    """Utility function that uses intall_name_tool to change oldReference to
    newReference in the machO file specified by fileName."""
    if VERBOSE:
        print("Redirecting load reference for ", end="")
        print(f"<{fileName}> {oldReference} -> {newReference}")
    original = os.stat(fileName).st_mode
    newMode = original | stat.S_IWUSR
    os.chmod(fileName, newMode)
    subprocess.call(
        ("install_name_tool", "-change", oldReference, newReference, fileName)
    )
    os.chmod(fileName, original)


def applyAdHocSignature(fileName: str):
    if platform.machine() != "arm64":
        return

    args = (
        "codesign",
        "--sign",
        "-",
        "--force",
        "--preserve-metadata=entitlements,requirements,flags,runtime",
        fileName,
    )
    if subprocess.call(args):
        # It may be a bug in Apple's codesign utility
        # The workaround is to copy the file to another inode, then move it
        # back erasing the previous file. The sign again.
        with tempfile.TemporaryDirectory() as dirName:
            tempName = os.path.join(dirName, os.path.basename(fileName))
            shutil.copy(fileName, tempName)
            shutil.move(tempName, fileName)
        subprocess.call(args)


class DarwinFileTracker:
    """Object to track the DarwinFiles that have been added during a freeze."""

    def __init__(self):
        # list of DarwinFile objects for files being copied into project
        self._copied_file_list: List[DarwinFile] = []

        # mapping of (build directory) target paths to DarwinFile objects
        self._darwin_file_for_build_path: Dict[Path, DarwinFile] = {}

        # mapping of (source location) paths to DarwinFile objects
        self._darwin_file_for_source_path: Dict[Path, DarwinFile] = {}

        # a cache of MachOReference objects pointing to a given source path
        self._reference_cache: Dict[Path, MachOReference] = {}

    def __iter__(self) -> Iterable[DarwinFile]:
        return iter(self._copied_file_list)

    def pathIsAlreadyCopiedTo(self, target_path: Path) -> bool:
        """Check if the given target_path has already has a file copied to
        it."""
        if target_path in self._darwin_file_for_build_path:
            return True
        return False

    def getDarwinFile(
        self, source_path: Path, target_path: Path
    ) -> DarwinFile:
        """Gets the DarwinFile for file copied from source_path to target_path.
        If either (i) nothing, or (ii) a different file has been copied to
        targetPath, raises a DarwinException."""
        # check that the target file came from the specified source
        targetDarwinFile: DarwinFile
        try:
            targetDarwinFile = self._darwin_file_for_build_path[target_path]
        except KeyError:
            raise DarwinException(
                f"File {target_path} already copied to, "
                "but no DarwinFile object found for it."
            ) from None
        real_source = source_path.resolve()
        target_real_source = targetDarwinFile.path.resolve()
        if real_source != target_real_source:
            # raise DarwinException(
            print(
                "*** WARNING ***\n"
                f"Attempting to copy two files to {target_path}\n"
                f"source 1: {targetDarwinFile.path} "
                f"(real: {target_real_source})\n"
                f"source 2: {source_path} (real: {real_source})\n"
                "(This may be caused by including modules in the zip file "
                "that rely on binary libraries with the same name.)"
                "\nUsing only source 1."
            )
        return targetDarwinFile

    def recordCopiedFile(self, target_path: Path, darwin_file: DarwinFile):
        """Record that a DarwinFile is being copied to a given path. If a
        file has been copied to that path, raise a DarwinException."""
        if self.pathIsAlreadyCopiedTo(target_path):
            raise DarwinException(
                "addFile() called with target_path already copied to "
                f"(target_path={target_path})"
            )

        self._copied_file_list.append(darwin_file)
        self._darwin_file_for_build_path[target_path] = darwin_file
        self._darwin_file_for_source_path[darwin_file.path] = darwin_file

    def cacheReferenceTo(self, source_path: Path, reference: MachOReference):
        self._reference_cache[source_path] = reference

    def getCachedReferenceTo(
        self, source_path: Path
    ) -> Optional[MachOReference]:
        return self._reference_cache.get(source_path)

    def findDarwinFileForFilename(self, filename: str) -> Optional[DarwinFile]:
        """Attempts to locate a copied DarwinFile with the specified filename
        and returns that. Otherwise returns None."""
        basename = Path(filename).name
        for file in self._copied_file_list:
            if file.path.name == basename:
                return file
        return None

    def finalizeReferences(self):
        """This function does a final pass through the references for all the
        copied DarwinFiles and attempts to clean up any remaining references
        that are not already marked as copied. It covers two cases where the
        reference might not be marked as copied:
        1) Files where _CopyFile was called without copyDependentFiles=True
           (in which the information would not have been added to the
            references at that time).
        2) Files with broken @rpath references. We try to fix that up here by
           seeing if the relevant file was located *anywhere* as part of the
           freeze process."""
        copied_file: DarwinFile
        reference: MachOReference
        for copied_file in self._copied_file_list:
            for reference in copied_file.getMachOReferenceList():
                if not reference.is_copied:
                    if reference.isResolved():
                        # if reference is resolve, simply check if the resolved
                        # path was otherwise copied and lookup the DarwinFile
                        # object.
                        target_path = reference.resolved_path.resolve()
                        if target_path in self._darwin_file_for_source_path:
                            reference.setTargetFile(
                                self._darwin_file_for_source_path[target_path]
                            )
                    else:
                        # if reference is not resolved, look through the copied
                        # files and try to find a candidate, and use it if
                        # found.
                        potential_target = self.findDarwinFileForFilename(
                            reference.raw_path
                        )
                        if potential_target is None:
                            # If we cannot find any likely candidate, fail.
                            print(
                                "\nERROR: Could not resolve RPath "
                                f"[{reference.raw_path}] in file "
                                f"[{copied_file.path}], and could "
                                "not find any likely intended reference."
                            )
                            copied_file.printFileInformation()
                            raise DarwinException(
                                f"finalizeReferences() failed to resolve path "
                                f"[{reference.raw_path}] in file "
                                f"[{copied_file.path}]."
                            )
                        print(
                            f"WARNING: In file [{copied_file.path}]"
                            f" guessing that {reference.raw_path} "
                            f"resolved to {potential_target.path}."
                        )
                        reference.resolved_path = potential_target.path
                        reference.setTargetFile(potential_target)



"""Module for the VersionInfo base class."""

import argparse
import json
import os
from pathlib import Path
from struct import calcsize, pack
from typing import List, Optional, Tuple, Union

from packaging.version import Version

try:
    from win32verstamp import stamp as version_stamp
except ImportError:
    version_stamp = None

try:
    from cx_Freeze import util
except ImportError:
    util = None

__all__ = ["VersionInfo"]

# types
CHAR = "c"
DWORD = "L"
WCHAR = "H"
WORD = "H"

VS_FFI_SIGNATURE = 0xFEEF04BD
VS_FFI_STRUCVERSION = 0x00010000
VS_FFI_FILEFLAGSMASK = 0x0000003F
VOS_NT_WINDOWS32 = 0x00040004

KEY_VERSION_INFO = "VS_VERSION_INFO"
KEY_STRING_FILE_INFO = "StringFileInfo"
KEY_STRING_TABLE = "040904E4"
KEY_VAR_FILE_INFO = "VarFileInfo"

# To disable the experimental feature in Windows:
# set CX_FREEZE_STAMP=pywin32
# pip install -U pywin32
if os.environ.get("CX_FREEZE_STAMP", "") == "pywin32":
    CX_FREEZE_STAMP = "pywin32"
else:
    CX_FREEZE_STAMP = "internal"


class Structure:
    """Abstract base class for structures in native byte order. Concrete
    structure and union types must be created by subclassing one of these
    types, and at least define a _fields class variable."""

    def __init__(self, *args):
        if not hasattr(self, "_fields"):
            self._fields: List[Tuple[str, str]] = []
        for i, (field, _) in enumerate(self._fields):
            setattr(self, field, args[i])

    def __str__(self):
        dump = json.dumps(self.as_dict(), indent=2)
        return self.__class__.__name__ + ": " + dump

    def as_dict(self):
        """Return the field values as dictionary."""
        fields = {}
        for fieldname, _ in self._fields:
            data = getattr(self, fieldname)
            if hasattr(data, "as_dict"):
                data = data.as_dict()
            elif isinstance(data, bytes):
                data = data.decode()
            fields[fieldname] = data
        return fields

    def to_buffer(self):
        """Return the field values to a buffer."""
        buffer = b""
        for fieldname, fmt in self._fields:
            data = getattr(self, fieldname)
            if hasattr(data, "to_buffer"):
                data = data.to_buffer()
            elif isinstance(data, str):
                data = data.encode("utf-16le")
            elif isinstance(fmt, str):
                data = pack(fmt, data)
            buffer += data
        return buffer


class VS_FIXEDFILEINFO(Structure):  # pylint: disable=invalid-name
    """Version information for a Win32 file."""

    _fields = [
        ("dwSignature", DWORD),
        ("dwStrucVersion", DWORD),
        ("dwFileVersionMS", DWORD),
        ("dwFileVersionLS", DWORD),
        ("dwProductVersionMS", DWORD),
        ("dwProductVersionLS", DWORD),
        ("dwFileFlagsMask", DWORD),
        ("dwFileFlags", DWORD),
        ("dwFileOS", DWORD),
        ("dwFileType", DWORD),
        ("dwFileSubtype", DWORD),
        ("dwFileDateMS", DWORD),
        ("dwFileDateLS", DWORD),
    ]


class String(Structure):
    """File version resource representation of the data."""

    def __init__(
        self, key: str, value: Optional[Union[int, str, Structure]] = None
    ):  # pylint: disable=W0231
        key = key + "\0"
        key_len = len(key)
        fields = [
            ("wLength", WORD),
            ("wValueLength", WORD),
            ("wType", WORD),
            ("szKey", WCHAR * key_len),
        ]
        key_len = calcsize(WCHAR) * key_len
        pad_len = (4 - ((calcsize(WORD) * 3 + key_len) & 3)) & 3
        if 0 < pad_len < 4:
            fields.append(("Padding", f"{pad_len}s"))
        value_len = 0
        value_type = 1
        value_size = 1
        if isinstance(value, int):
            value_len = calcsize(DWORD)
            value_type = 0
            fields.append(("Value", DWORD))
        elif isinstance(value, str):
            value = value + "\0"
            value_len = len(value)
            value_size = calcsize(WCHAR)
            fields.append(("Value", WCHAR * value_len))
        elif hasattr(value, "wLength"):  # instance of String
            value_len = value.wLength
            fields.append(("Value", type(value)))
        elif isinstance(value, Structure):
            value_len = calcsize("".join([f[1] for f in value._fields]))
            value_type = 0
            fields.append(("Value", type(value)))

        self._fields = fields
        # pylint: disable-next=invalid-name
        self.wValueLength = value_len
        # pylint: disable-next=invalid-name
        self.wType = value_type
        # pylint: disable-next=invalid-name
        self.szKey = key
        # pylint: disable-next=invalid-name
        self.Padding = b"\0" * pad_len
        # pylint: disable-next=invalid-name
        self.Value = value
        # pylint: disable-next=invalid-name
        self.wLength = (
            calcsize(WORD) * 3 + key_len + pad_len + value_size * value_len
        )
        self._children = 0

    def children(self, value: "String"):
        """Represents the child String object."""
        pad_len = 4 - (self.wLength & 3)
        if 0 < pad_len < 4:
            field = f"Padding{self._children}"
            self._fields.append((field, f"{pad_len}s"))
            setattr(self, field, b"\0" * pad_len)
            self.wLength += calcsize(CHAR) * pad_len
        field = f"Children{self._children}"
        self._fields.append((field, type(value)))
        setattr(self, field, value)
        self._children += 1
        self.wLength += value.wLength


class VersionInfo:
    """Organizes the version information (resource) data of a file."""

    def __init__(
        self,
        version: str,
        internal_name: Optional[str] = None,
        original_filename: Optional[str] = None,
        comments: Optional[str] = None,
        company: Optional[str] = None,
        description: Optional[str] = None,
        copyright: Optional[str] = None,  # pylint: disable=W0622
        trademarks: Optional[str] = None,
        product: Optional[str] = None,
        dll: Optional[bool] = None,
        debug: Optional[bool] = None,
        verbose: bool = True,
    ):
        self.version: Version = Version(version)
        self.internal_name: Optional[str] = internal_name
        self.original_filename: Optional[str] = original_filename
        self.comments: Optional[str] = comments
        self.company: Optional[str] = company
        self.description: Optional[str] = description
        self.copyright: Optional[str] = copyright
        self.trademarks: Optional[str] = trademarks
        self.product: Optional[str] = product
        self.dll: Optional[bool] = dll
        self.debug: Optional[bool] = debug
        self.verbose: bool = verbose

    def stamp(self, path: Union[str, Path]) -> None:
        """Stamp a Win32 binary with version information."""
        if isinstance(path, str):
            path = Path(path)
        if not path.is_file():
            raise FileNotFoundError(path)

        if CX_FREEZE_STAMP == "pywin32":
            if version_stamp is None:
                raise RuntimeError("install pywin32 extensions first")

            version_stamp(str(path), self)
            return

        # internal
        string_version_info = self.version_info(path)
        if CX_FREEZE_STAMP == "internal":
            if util is None:
                raise RuntimeError("cx_Freeze.util extensions not found")
            handle = util.BeginUpdateResource(path, 0)
            util.UpdateResource(handle, 16, 1, string_version_info.to_buffer())
            util.EndUpdateResource(handle, 0)

        if self.verbose:
            print("Stamped:", path)

    def version_info(self, path: Path) -> String:
        """Returns the String version info used to stamp the version."""
        major = self.version.major
        minor = self.version.minor
        micro = self.version.micro
        build = 0
        file_flags = 0
        if self.debug is None or path.stem.lower().endswith("_d"):
            file_flags += 1
        if self.version.is_devrelease:
            file_flags += 8
            build = self.version.dev
        elif self.version.is_prerelease:
            file_flags += 2
            build = self.version.pre[1]
        elif self.version.is_postrelease:
            file_flags += 0x20
            build = self.version.post

        data = {
            "Comments": self.comments or "",
            "CompanyName": self.company or "",
            "FileDescription": self.description or "",
            "FileVersion": self.version.base_version,
            "InternalName": self.internal_name or path.name,
            "LegalCopyright": self.copyright or "",
            "LegalTrademarks": self.trademarks or "",
            "OriginalFilename": self.original_filename or path.name,
            "ProductName": self.product or "",
            "ProductVersion": str(self.version),
        }
        is_dll = self.dll
        if is_dll is None:
            is_dll = path.suffix.lower() in (".dll", ".pyd")
        fixed_file_info = VS_FIXEDFILEINFO(
            VS_FFI_SIGNATURE,
            VS_FFI_STRUCVERSION,
            (major << 16) | minor,
            (micro << 16) | build,
            (major << 16) | minor,
            (micro << 16) | build,
            VS_FFI_FILEFLAGSMASK,
            file_flags,
            VOS_NT_WINDOWS32,
            2 if is_dll else 1,  # VFT_DLL or VFT_APP
            0,
            0,
            0,
        )

        # string table with its children
        string_table = String(KEY_STRING_TABLE)
        for key, value in data.items():
            string_table.children(String(key, value))

        # create string file info and add string table as child
        string_file_info = String(KEY_STRING_FILE_INFO)
        string_file_info.children(string_table)

        # var file info has a child
        var_file_info = String(KEY_VAR_FILE_INFO)
        var_file_info.children(String("Translation", 0x04E40409))  # 0x409,1252

        # VS_VERSION_INFO is the first key and has two children
        string_version_info = String(KEY_VERSION_INFO, fixed_file_info)
        string_version_info.children(string_file_info)
        string_version_info.children(var_file_info)
        return string_version_info


if __name__ == "__main__":
    # simple test
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "filename",
        nargs="?",
        metavar="NAME",
        help="the name of the file (.dll, .pyd or .exe) to test version stamp",
    )
    parser.add_argument(
        "--version",
        action="store",
        dest="version",
        default="0.1",
        help="version to set as test",
    )
    parser.add_argument(
        "--dict",
        action="store_true",
        dest="as_dict",
        help="show version info as dict",
    )
    parser.add_argument(
        "--raw",
        action="store_true",
        dest="as_raw",
        help="show version info as raw bytes",
    )
    test_args = parser.parse_args()
    if test_args.filename is None:
        parser.error("filename must be specified")
    else:
        test_filename = Path(test_args.filename)

    test_version = VersionInfo(
        test_args.version,
        comments="cx_Freeze comments",
        description="cx_Freeze description",
        company="cx_Freeze company",
        product="cx_Freeze product",
        copyright="(c) 2022, cx_Freeze",
        trademarks="cx_Freeze (TM)",
    )

    if test_args.as_dict:
        print(test_version.version_info(test_filename))
    if test_args.as_raw:
        print(test_version.version_info(test_filename).to_buffer().hex(":"))
    test_version.stamp(test_filename)



"""cxfreeze command line tool."""

import argparse
import os
import sys
import sysconfig

from . import __version__
from .common import normalize_to_list
from .executable import Executable
from .freezer import Freezer

__all__ = ["main"]

DESCRIPTION = """
Freeze a Python script and all of its referenced modules to a base \
executable which can then be distributed without requiring a Python \
installation.
"""

VERSION = f"""
%(prog)s {__version__}
Copyright (c) 2020-2022 Marcelo Duarte. All rights reserved.
Copyright (c) 2007-2020 Anthony Tuininga. All rights reserved.
Copyright (c) 2001-2006 Computronix Corporation. All rights reserved.
"""


def prepare_parser():
    """Helper function to parse the arguments."""

    parser = argparse.ArgumentParser(prog="cxfreeze", epilog=VERSION)
    parser.add_argument("--version", action="version", version=VERSION)
    # Executable parameters
    parser.add_argument(
        "script",
        nargs="?",
        metavar="SCRIPT",
        help="the name of the file containing the script which is to be "
        "frozen",
    )
    parser.add_argument(
        "--init-script",
        dest="init_script",
        metavar="NAME",
        help="script which will be executed upon startup; if the name of the "
        "file is not an absolute file name, the subdirectory initscripts "
        "(rooted in the directory in which the cx_Freeze package is found) "
        "will be searched for a file matching the name",
    )
    parser.add_argument(
        "--base-name",
        dest="base_name",
        metavar="NAME",
        help="the filename of the base executable; if a name is given without "
        "an absolute path, the subdirectory bases (rooted in the directory in "
        "which the freezer is found) will be searched for a file matching the "
        "name",
    )
    parser.add_argument(
        "--target-name",
        dest="target_name",
        metavar="NAME",
        help="the name of the target executable; the default value is the "
        "name of the script; the extension is optional (automatically added "
        "on Windows); support for names with version; if specified a "
        "pathname, raise an error",
    )
    parser.add_argument(
        "--icon",
        dest="icon",
        metavar="NAME",
        help="name of icon which should be included in the executable itself "
        "on Windows or placed in the target directory for other platforms "
        "(ignored by Python app from Microsoft Store)",
    )
    parser.add_argument(
        "--manifest",
        dest="manifest",
        metavar="NAME",
        help="name of manifest which should be included in the executable "
        "itself (Windows only - ignored by Python app from Microsoft Store)",
    )
    parser.add_argument(
        "--uac-admin",
        action="store_true",
        dest="uac_admin",
        help="creates a manifest for an application that will request eleva"
        "tion (Windows only - ignored by Python app from Microsoft Store)",
    )
    parser.add_argument(
        "--shortcut-name",
        dest="shortcut_name",
        metavar="NAME",
        help="the name to give a shortcut for the executable when included in "
        "an MSI package (Windows only)",
    )
    parser.add_argument(
        "--shortcut-dir",
        dest="shortcut_dir",
        metavar="DIR",
        help="the directory in which to place the shortcut when being instal"
        "led by an MSI package; see the MSI Shortcut table documentation for "
        "more information on what values can be placed here (Windows only)",
    )
    parser.add_argument(
        "--copyright",
        dest="copyright",
        help="the copyright value to include in the version resource "
        "associated with executable (Windows only)",
    )
    parser.add_argument(
        "--trademarks",
        dest="trademarks",
        help="the trademarks value to include in the version resource "
        "associated with the executable (Windows only)",
    )
    # Freezer parameters
    platform = sysconfig.get_platform()
    python_version = sysconfig.get_python_version()
    dir_name = f"exe.{platform}-{python_version}"
    parser.add_argument(
        "--target-dir",
        "--install-dir",
        dest="target_dir",
        default=os.path.abspath(os.path.join("build", dir_name)),
        metavar="DIR",
        help="directory for built executable and dependent files"
        " (default: %(default)s)",
    )
    parser.add_argument(
        "-O",
        action="count",
        default=0,
        dest="optimize_flag",
        help="optimize generated bytecode as per PYTHONOPTIMIZE; "
        "use -OO in order to remove doc strings",
    )
    parser.add_argument(
        "--excludes",
        "--exclude-modules",
        dest="excludes",
        metavar="NAMES",
        help="comma-separated list of modules to exclude",
    )
    parser.add_argument(
        "--includes",
        "--include-modules",
        dest="includes",
        metavar="NAMES",
        help="comma-separated list of modules to include",
    )
    parser.add_argument(
        "--packages",
        dest="packages",
        metavar="NAMES",
        help="comma-separated list of packages to include, "
        "which includes all submodules in the package",
    )
    parser.add_argument(
        "--replace-paths",
        dest="replace_paths",
        metavar="DIRECTIVES",
        help="comma-separated list of paths to replace in the code object of "
        "included modules, using the form <search>=<replace>; search can be * "
        "which means all paths not already specified, leaving just the "
        "relative path to the module; multiple values are separated by the "
        "standard path separator",
    )
    parser.add_argument(
        "--default-path",
        action="append",
        dest="default_path",
        metavar="DIRS",
        help="list of paths separated by the standard path separator for the "
        "platform which will be used to initialize sys.path prior to running "
        "the module finder",
    )
    parser.add_argument(
        "--include-path",
        action="append",
        dest="include_path",
        metavar="DIRS",
        help="list of paths separated by the standard path separator for the "
        "platform which will be used to modify sys.path prior to running the "
        "module finder",
    )
    parser.add_argument(
        "-c",
        "--compress",
        action="store_true",
        dest="compress",
        help="compress byte code in zip files",
    )
    # TODO: "constants"
    parser.add_argument(
        "--bin-includes",
        dest="bin_includes",
        help="comma-separated list of files to include when determining "
        "dependencies of binary files that would normally be excluded, using "
        "first the full file name, then just the base file name, then the "
        "file name without any version numbers (the version numbers that "
        "normally follow the shared object extension are stripped prior to "
        "performing the comparison)",
    )
    parser.add_argument(
        "--bin-excludes",
        dest="bin_excludes",
        help="comma-separated list of files to exclude when determining "
        "dependencies of binary files that would normally be included, using "
        "first the full file name, then just the base file name, then the "
        "file name without any version numbers (the version numbers that "
        "normally follow the shared object extension are stripped prior to "
        "performing the comparison)",
    )
    parser.add_argument(
        "--bin-path-includes",
        dest="bin_path_includes",
        help="comma-separated list of paths from which to include files when "
        "determining dependencies of binary files",
    )
    parser.add_argument(
        "--bin-path-excludes",
        dest="bin_path_excludes",
        help="comma-separated list of paths from which to exclude files when "
        "determining dependencies of binary files",
    )
    parser.add_argument(
        "--include-files",
        dest="include_files",
        metavar="NAMES",
        help="comma-separated list of additional files to include in "
        "distribution",
    )
    parser.add_argument(
        "-z",
        "--zip-include",
        dest="zip_includes",
        action="append",
        default=[],
        metavar="SPEC",
        help="additional file to include in zip file or a specification of "
        "the form name=arcname which will specify the archive name to use; "
        "multiple --zip-include arguments can be used",
    )
    parser.add_argument(
        "--zip-include-packages",
        dest="zip_include_packages",
        metavar="NAMES",
        help="comma-separated list of packages which should be included in "
        "the zip file; the default is for all packages to be placed in the "
        "file system, not the zip file; those packages which are known to "
        "work well inside a zip file can be included if desired; use * to "
        "specify that all packages should be included in the zip file",
    )
    parser.add_argument(
        "--zip-exclude-packages",
        dest="zip_exclude_packages",
        default="*",
        metavar="NAMES",
        help="comma-separated list of packages which should be excluded from "
        "the zip file and placed in the file system instead; the default is "
        "for all packages to be placed in the file system since a number of pa"
        "ckages assume that is where they are found and will fail when placed "
        "in a zip file; use * to specify that all packages should be placed "
        "in the file system and excluded from the zip file (the default)",
    )
    parser.add_argument(
        "-s",
        "--silent",
        action="store_true",
        dest="silent",
        help="suppress all output except warnings and errors",
    )
    parser.add_argument(
        "--include-msvcr",
        action="store_true",
        dest="include_msvcr",
        help="include the Microsoft Visual C runtime files",
    )
    # remove the initial "usage: " of format_usage()
    parser.usage = parser.format_usage()[len("usage: ") :] + DESCRIPTION
    return parser


def parse_command_line(parser):
    """Helper function to parse the command line."""

    args = parser.parse_args()
    if args.script is None and args.includes is None:
        parser.error("script or a list of modules must be specified")
    if args.script is None and args.target_name is None:
        parser.error("script or a target name must be specified")
    args.includes = normalize_to_list(args.includes)
    args.excludes = normalize_to_list(args.excludes)
    args.packages = normalize_to_list(args.packages)
    replace_paths = []
    if args.replace_paths:
        for directive in args.replace_paths.split(os.pathsep):
            from_path, replacement = directive.split("=")
            replace_paths.append((from_path, replacement))
    args.replace_paths = replace_paths
    if args.default_path is not None:
        sys.path = [
            p for mp in args.default_path for p in mp.split(os.pathsep)
        ]
    if args.include_path is not None:
        paths = [p for mp in args.include_path for p in mp.split(os.pathsep)]
        sys.path = paths + sys.path
    if args.script is not None:
        sys.path.insert(0, os.path.dirname(args.script))
    args.bin_includes = normalize_to_list(args.bin_includes)
    args.bin_excludes = normalize_to_list(args.bin_excludes)
    args.bin_path_includes = normalize_to_list(args.bin_path_includes)
    args.bin_path_excludes = normalize_to_list(args.bin_path_excludes)
    args.include_files = normalize_to_list(args.include_files)
    zip_includes = []
    if args.zip_includes:
        for spec in args.zip_includes:
            if "=" in spec:
                zip_includes.append(spec.split("=", 1))
            else:
                zip_includes.append(spec)
    args.zip_includes = zip_includes
    args.zip_include_packages = normalize_to_list(args.zip_include_packages)
    args.zip_exclude_packages = normalize_to_list(args.zip_exclude_packages)
    return args


def main():
    """Entry point for cxfreeze command line tool."""

    args = parse_command_line(prepare_parser())
    executables = [
        Executable(
            args.script,
            args.init_script,
            args.base_name,
            args.target_name,
            args.icon,
            args.shortcut_name,
            args.shortcut_dir,
            args.copyright,
            args.trademarks,
            args.manifest,
            args.uac_admin,
        )
    ]
    freezer: Freezer = Freezer(
        executables,
        excludes=args.excludes,
        includes=args.includes,
        packages=args.packages,
        replacePaths=args.replace_paths,
        compress=args.compress,
        optimizeFlag=args.optimize_flag,
        path=None,
        targetDir=args.target_dir,
        binIncludes=args.bin_includes,
        binExcludes=args.bin_excludes,
        binPathIncludes=args.bin_path_includes,
        binPathExcludes=args.bin_path_excludes,
        includeFiles=args.include_files,
        zipIncludes=args.zip_includes,
        zipIncludePackages=args.zip_include_packages,
        zipExcludePackages=args.zip_exclude_packages,
        silent=args.silent,
        includeMSVCR=args.include_msvcr,
    )
    freezer.Freeze()



"""Extend setuptools.sandbox.run_setup to work with cx_Freeze setup()."""

import contextlib
import sys
from pathlib import Path
from typing import Union

from setuptools.sandbox import (
    DirectorySandbox,
    override_temp,
    pushd,
    save_argv,
    save_path,
)

__all__ = ["run_setup"]


def run_setup(setup_script: Union[str, Path], args):
    """Run a cx_Freeze setup script, sandboxed in its directory."""

    fullpath = Path(setup_script).resolve()
    setup_dir = str(fullpath.parent)
    setup_script = str(fullpath)
    with _setup_context(setup_dir):
        try:
            sys.argv[:] = [setup_script] + list(args)
            sys.path.insert(0, setup_dir)
            with DirectorySandbox(setup_dir):
                env = {"__file__": setup_script, "__name__": "__main__"}
                _execfile(setup_script, env)
        except SystemExit as exc:
            if exc.args and exc.args[0]:
                raise
            # Normal exit, just return


@contextlib.contextmanager
def _setup_context(setup_dir: str):
    """Save and prepare de environment to run_setup."""

    temp_dir = str(Path(setup_dir, "temp"))
    with save_path():
        with save_argv():
            with override_temp(temp_dir):
                with pushd(setup_dir):
                    # ensure cx_Freeze commands are available
                    # __import__("cx_Freeze")
                    yield


# pylint: disable-next=redefined-builtin
def _execfile(filename: str, globals, locals=None):
    """Python 3 implementation of execfile."""

    script = Path(filename).read_bytes()
    if locals is None:
        locals = globals
    code = compile(script, filename, "exec")
    exec(code, globals, locals)  # pylint: disable=exec-used



"""Extends setuptools 'install' command."""

import contextlib
import sys
import warnings

from setuptools.command.install import install as _install

__all__ = ["Install"]


@contextlib.contextmanager
def suppress_known_deprecation():  # pylint: disable=missing-function-docstring
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", "setup.py install is deprecated")
        yield


# pylint: disable=attribute-defined-outside-init
class Install(_install):
    """Install everything from build directory."""

    user_options = _install.user_options + [
        ("install-exe=", None, "installation directory for executables")
    ]

    def expand_dirs(self):
        _install.expand_dirs(self)
        self._expand_attrs(["install_exe"])

    def get_sub_commands(self):
        sub_commands = _install.get_sub_commands(self)
        if self.distribution.executables:
            sub_commands.remove("install_egg_info")
            sub_commands.remove("install_scripts")
            sub_commands.append("install_exe")
        return sub_commands

    def initialize_options(self):
        with suppress_known_deprecation():
            _install.initialize_options(self)
        self.install_exe = None

    def finalize_options(self):
        if self.prefix is None and sys.platform == "win32":
            winreg = __import__("winreg")
            key = winreg.OpenKey(
                winreg.HKEY_LOCAL_MACHINE,
                r"Software\Microsoft\Windows\CurrentVersion",
            )
            prefix = str(winreg.QueryValueEx(key, "ProgramFilesDir")[0])
            metadata = self.distribution.metadata
            self.prefix = f"{prefix}/{metadata.name}"
        _install.finalize_options(self)
        self.convert_paths("exe")
        if self.root is not None:
            self.change_roots("exe")

    def select_scheme(self, name):
        _install.select_scheme(self, name)
        if self.install_exe is None:
            if sys.platform == "win32":
                self.install_exe = "$base"
            else:
                metadata = self.distribution.metadata
                dir_name = f"{metadata.name}-{metadata.version}"
                self.install_exe = f"$base/lib/{dir_name}"



"""Implements the 'build_exe' command."""

import logging
import os
import sys
import warnings

from setuptools import Command
from setuptools.errors import SetupError

from ..common import normalize_to_list
from ..freezer import Freezer
from ..module import ConstantsModule

__all__ = ["BuildEXE"]


# pylint: disable=attribute-defined-outside-init,missing-function-docstring
class BuildEXE(Command):
    """Build executables from Python scripts."""

    description = "build executables from Python scripts"
    user_options = [
        (
            "build-exe=",
            "b",
            "directory for built executables and dependent files",
        ),
        (
            "optimize=",
            "O",
            'optimization level: -O1 for "python -O", '
            '-O2 for "python -OO" and -O0 to disable [default: -O0]',
        ),
        ("excludes=", "e", "comma-separated list of modules to exclude"),
        ("includes=", "i", "comma-separated list of modules to include"),
        (
            "packages=",
            "p",
            "comma-separated list of packages to include, "
            "which includes all submodules in the package",
        ),
        ("namespace-packages=", None, "[DEPRECATED]"),
        (
            "replace-paths=",
            None,
            "comma-separated list of paths to replace in included modules, "
            "using the form <search>=<replace>",
        ),
        ("path=", None, "comma-separated list of paths to search"),
        ("no-compress", None, "create a zipfile with no compression"),
        ("constants=", None, "comma-separated list of constants to include"),
        (
            "bin-includes",
            None,
            "list of files to include when determining "
            "dependencies of binary files that would normally be excluded",
        ),
        (
            "bin-excludes",
            None,
            "list of files to exclude when determining "
            "dependencies of binary files that would normally be included",
        ),
        (
            "bin-path-includes",
            None,
            "list of paths from which to include files when determining "
            "dependencies of binary files",
        ),
        (
            "bin-path-excludes",
            None,
            "list of paths from which to exclude files when determining "
            "dependencies of binary files",
        ),
        (
            "include-files=",
            "f",
            "list of tuples of additional files to include in distribution",
        ),
        (
            "zip-includes=",
            None,
            "list of tuples of additional files to include in zip file",
        ),
        (
            "zip-include-packages=",
            None,
            "comma-separated list of packages to include in the zip file "
            "(or * for all) [default: none]",
        ),
        (
            "zip-exclude-packages=",
            None,
            "comma-separated list of packages to exclude from the zip file "
            "and place in the file system instead (or * for all) "
            "[default: *]",
        ),
        (
            "silent",
            "s",
            "suppress all output except warnings "
            "(equivalent to --silent-level=1)",
        ),
        (
            "silent-level=",
            None,
            "suppress output from build_exe command."
            " level 0: get all messages; [default]"
            " level 1: suppress information messages, but still get warnings;"
            " (equivalent to --silent)"
            " level 2: suppress missing missing-module warnings"
            " level 3: suppress all warning messages",
        ),
        (
            "include-msvcr=",
            None,
            "include the Microsoft Visual C runtime files",
        ),
    ]
    boolean_options = ["no-compress", "include_msvcr", "silent"]

    def add_to_path(self, name):
        source_dir = getattr(self, name.lower())
        if source_dir is not None:
            sys.path.insert(0, source_dir)

    def build_extension(self, name, module_name=None):
        # XXX: This method, add_to_path and set_source_location can be deleted?
        if module_name is None:
            module_name = name
        source_dir = getattr(self, name.lower())
        if source_dir is None:
            return None
        orig_dir = os.getcwd()
        script_args = ["build"]
        command = self.distribution.get_command_obj("build")
        if command.compiler is not None:
            script_args.append(f"--compiler={command.compiler}")
        os.chdir(source_dir)
        logging.info("building '%s' extension in '%s'", name, source_dir)
        distutils_core = __import__("distutils.core", fromlist=["run_setup"])
        distribution = distutils_core.run_setup("setup.py", script_args)
        ext_modules = distribution.ext_modules
        modules = [m for m in ext_modules if m.name == module_name]
        if not modules:
            raise SetupError(
                "no module named '{module_name}' in '{source_dir}'"
            )
        command = distribution.get_command_obj("build_ext")
        command.ensure_finalized()
        if command.compiler is None:
            command.run()
        else:
            command.build_extensions()
        dir_name = os.path.join(source_dir, command.build_lib)
        os.chdir(orig_dir)
        if dir_name not in sys.path:
            sys.path.insert(0, dir_name)
        return os.path.join(
            source_dir,
            command.build_lib,
            command.get_ext_filename(module_name),
        )

    def initialize_options(self):
        self.list_options = [
            "excludes",
            "includes",
            "packages",
            "replace_paths",
            "constants",
            "bin_excludes",
            "bin_includes",
            "bin_path_includes",
            "bin_path_excludes",
            "include_files",
            "zip_includes",
            "zip_include_packages",
            "zip_exclude_packages",
            # DEPRECATED
            "namespace_packages",
        ]

        for option in self.list_options:
            setattr(self, option, [])

        self.zip_exclude_packages = ["*"]
        self.optimize = 0
        self.build_exe = None
        self.no_compress = False
        self.path = None
        self.include_msvcr = None
        self.silent = None
        self.silent_level = None

    def finalize_options(self):
        if self.build_exe is None:
            self.set_undefined_options("build", ("build_exe", "build_exe"))
        else:
            self.get_finalized_command("build").build_exe = self.build_exe
        self.optimize = int(self.optimize)

        # the degree of silencing, set from either the silent or silent-level
        # option, as appropriate
        self.silent_setting = 0
        if self.silent is not None and self.silent:
            self.silent_setting = 1

        if self.silent_level is None:
            pass
        elif self.silent_level is False:
            self.silent_setting = 0
        elif self.silent_level is True:
            self.silent_setting = 1
        elif isinstance(self.silent_level, int):
            self.silent_setting = self.silent_level
        elif isinstance(self.silent_level, str):
            try:
                self.silent_setting = int(self.silent_level)
            except ValueError:
                self.silent_setting = 1
        else:
            self.silent_setting = 1

        # Make sure all options of multiple values are lists
        for option in self.list_options:
            setattr(self, option, normalize_to_list(getattr(self, option)))

    def run(self):
        metadata = self.distribution.metadata
        constants_module = ConstantsModule(
            metadata.version, constants=self.constants
        )
        if self.namespace_packages:
            warnings.warn(
                "namespace-packages is obsolete and will be removed in the "
                "next version"
            )

        freezer: Freezer = Freezer(
            self.distribution.executables,
            constants_module,
            self.includes,
            self.excludes,
            self.packages,
            self.replace_paths,
            (not self.no_compress),
            self.optimize,
            self.path,
            self.build_exe,
            binIncludes=self.bin_includes,
            binExcludes=self.bin_excludes,
            binPathIncludes=self.bin_path_includes,
            binPathExcludes=self.bin_path_excludes,
            includeFiles=self.include_files,
            zipIncludes=self.zip_includes,
            zipIncludePackages=self.zip_include_packages,
            zipExcludePackages=self.zip_exclude_packages,
            silent=self.silent_setting,
            metadata=metadata,
            includeMSVCR=self.include_msvcr,
        )

        # keep freezer around so that its data case be used in bdist_mac phase
        self.freezer = freezer
        freezer.Freeze()

    def set_source_location(self, name, *pathParts):
        env_name = f"{name.upper()}_BASE"
        attr_name = name.lower()
        source_dir = getattr(self, attr_name)
        if source_dir is None:
            base_dir = os.environ.get(env_name)
            if base_dir is None:
                return
            source_dir = os.path.join(base_dir, *pathParts)
            if os.path.isdir(source_dir):
                setattr(self, attr_name, source_dir)



"""Internal compatible module."""
import sys

__all__ = ["cached_property", "importlib_metadata"]

if sys.version_info >= (3, 10):
    import importlib.metadata as importlib_metadata
else:
    try:
        import importlib_metadata
    except ImportError:
        import importlib.metadata as importlib_metadata


try:
    from functools import cached_property
except ImportError:

    class cached_property:  # pylint: disable=invalid-name
        """Transform a method of a class into a property whose value is
        computed once and then cached as a normal attribute for the life of
        the instance."""

        def __init__(self, func):
            self.func = func
            self.__doc__ = func.__doc__

        def __get__(self, instance, owner=None):
            if instance is None:
                return self

            value = instance.__dict__[self.func.__name__] = self.func(instance)
            return value



"""The core class for freezing scripts into executables."""

import marshal
import os
import shutil
import stat
import struct
import sys
import sysconfig
import time
from abc import abstractmethod
from importlib.util import MAGIC_NUMBER
from pathlib import Path
from typing import List, Optional, Set, Tuple, Union
from zipfile import ZIP_DEFLATED, ZIP_STORED, PyZipFile, ZipInfo

from .common import (
    IncludesList,
    InternalIncludesList,
    get_resource_file_path,
    process_path_specs,
)
from .dist import DistributionMetadata
from .exception import ConfigError
from .executable import Executable
from .finder import ModuleFinder
from .module import ConstantsModule, Module
from .parser import ELFParser, Parser, PEParser

DARWIN = sys.platform == "darwin"
MINGW = sysconfig.get_platform().startswith("mingw")
WIN32 = sys.platform == "win32"

if WIN32:
    from . import winmsvcr
    from .winversioninfo import VersionInfo

    try:
        from .util import AddIcon, GetSystemDir, GetWindowsDir, UpdateCheckSum
    except ImportError:
        pass
elif DARWIN:
    from .darwintools import DarwinFile, DarwinFileTracker, MachOReference

__all__ = ["ConfigError", "ConstantsModule", "Executable", "Freezer"]


class Freezer:
    def __new__(cls, *args, **kwargs):
        # create instance of appropriate sub-class, depending on the platform.
        instance: Freezer
        if WIN32:
            instance = super().__new__(WinFreezer)
        elif DARWIN:
            instance = super().__new__(DarwinFreezer)
        else:  # assume any other platform would be handled by LinuxFreezer
            instance = super().__new__(LinuxFreezer)
        return instance

    def __init__(
        self,
        executables: List[Executable],
        constantsModule: Optional[ConstantsModule] = None,
        includes: Optional[List[str]] = None,
        excludes: Optional[List[str]] = None,
        packages: Optional[List[str]] = None,
        replacePaths: Optional[List[str]] = None,
        compress: bool = True,
        optimizeFlag: int = 0,
        path: Optional[List[str]] = None,
        targetDir: Optional[Union[str, Path]] = None,
        binIncludes: Optional[List[str]] = None,
        binExcludes: Optional[List[str]] = None,
        binPathIncludes: Optional[List[str]] = None,
        binPathExcludes: Optional[List[str]] = None,
        includeFiles: Optional[IncludesList] = None,
        zipIncludes: Optional[IncludesList] = None,
        silent: Union[bool, int] = 0,
        metadata: Optional[DistributionMetadata] = None,
        includeMSVCR: bool = False,
        zipIncludePackages: Optional[List[str]] = None,
        zipExcludePackages: Optional[List[str]] = None,
    ):
        self.executables: List[Executable] = list(executables)
        if constantsModule is None:
            constantsModule = ConstantsModule()
        self.constants_module: ConstantsModule = constantsModule
        self.includes: List[str] = list(includes or [])
        self.excludes: List[str] = list(excludes or [])
        self.packages: Set[str] = set(packages or [])
        self.replacePaths: List[str] = list(replacePaths or [])
        self.compress = True if compress is None else compress
        self.optimize_flag: int = optimizeFlag
        self.path: Optional[List[str]] = path
        self.include_msvcr: bool = includeMSVCR
        self.targetdir = targetDir
        self.bin_includes: Optional[List[str]] = binIncludes
        self.bin_excludes: Optional[List[str]] = binExcludes
        self.bin_path_includes: Optional[List[str]] = binPathIncludes
        self.bin_path_excludes: Optional[List[str]] = binPathExcludes
        self.include_files: InternalIncludesList = process_path_specs(
            includeFiles
        )
        self.zip_includes: InternalIncludesList = process_path_specs(
            zipIncludes
        )
        if isinstance(silent, bool):
            if silent:
                self.silent = 1
            else:
                self.silent = 0
        else:
            self.silent = silent
        self.metadata: Optional[DistributionMetadata] = metadata
        self.zipIncludePackages: Optional[List[str]] = zipIncludePackages
        self.zipExcludePackages: Optional[List[str]] = zipExcludePackages
        self._verify_configuration()

    @property
    def targetdir(self) -> Path:
        return self._targetdir

    @targetdir.setter
    def targetdir(self, path: Optional[Union[str, Path]]):
        if path is None:
            platform = sysconfig.get_platform()
            python_version = sysconfig.get_python_version()
            path = f"build/exe.{platform}-{python_version}"
        path = Path(path).resolve()
        if path.is_dir():
            # starts in a clean directory

            def onerror(*args):
                raise ConfigError("the build directory cannot be cleaned")

            shutil.rmtree(path, onerror=onerror)

        self._targetdir: Path = path

    def _add_resources(self, exe: Executable) -> None:
        """Add resources for an executable, platform dependent."""
        # Copy icon into application. (Overridden on Windows)
        if exe.icon is None:
            return
        target_icon = self.targetdir / exe.icon.name
        self._copy_file(exe.icon, target_icon, copy_dependent_files=False)

    def _copy_file(
        self,
        source: Path,
        target: Path,
        copy_dependent_files: bool,
        include_mode: bool = False,
    ):
        if not self._should_copy_file(source):
            return

        # handle pre-copy tasks, normally on the target path
        source, target = self._pre_copy_hook(source, target)

        if target in self.files_copied:
            return
        if source == target:
            return
        self._create_directory(target.parent)
        if self.silent < 1:
            print(f"copying {source} -> {target}")
        shutil.copyfile(source, target)
        shutil.copystat(source, target)
        if include_mode:
            shutil.copymode(source, target)
        self.files_copied.add(target)

        # handle post-copy tasks, including copying dependencies
        self._post_copy_hook(
            source,
            target,
            copy_dependent_files=copy_dependent_files,
            include_mode=include_mode,
        )

    @abstractmethod
    def _pre_copy_hook(self, source: Path, target: Path) -> Tuple[Path, Path]:
        """Prepare the source and target paths."""

    @abstractmethod
    def _post_copy_hook(
        self,
        source: Path,
        target: Path,
        copy_dependent_files: bool,
        include_mode: bool = False,
    ):
        """Post-copy task."""

    def _create_directory(self, path: Union[str, Path]):
        if isinstance(path, str):
            path = Path(path)
        if not path.is_dir():
            if self.silent < 1:
                print(f"creating directory {path}")
            path.mkdir(parents=True, exist_ok=True)

    def _freeze_executable(self, exe: Executable) -> None:
        finder: ModuleFinder = self.finder
        finder.include_file_as_module(exe.main_script, exe.main_module_name)
        finder.include_file_as_module(exe.init_script, exe.init_module_name)
        finder.include_file_as_module(
            get_resource_file_path("initscripts", "__startup__", ".py")
        )

        # Ensure the copy of default python libraries
        dependent_files = self.get_dependent_files(exe.base)
        if not dependent_files:
            dependent_files = self.get_dependent_files(Path(sys.executable))
        python_libs = tuple(self._default_bin_includes())
        python_dirs = {Path(sys.base_exec_prefix), Path(sys.exec_prefix)}
        python_dirs.add(Path(sysconfig.get_config_var("srcdir")))  # Linux
        for file in dependent_files:
            python_dirs.add(file.parent)
        for name in python_libs:
            for python_dir in python_dirs:
                source = python_dir / name
                if source.is_file():
                    dependent_files.add(source)
                    break
        if not dependent_files:
            if self.silent < 3:
                print("WARNING: shared libraries not found:", python_libs)

        # Search the C runtimes, using the directory of the python libraries
        # and the directories of the base executable
        self._platform_add_extra_dependencies(dependent_files)

        for source in dependent_files:
            # Store dynamic libraries in appropriate location for platform.
            self._copy_top_dependency(source)
            # Once copied, it should be deleted from the list to ensure
            # it will not be copied again.
            name = Path(source.name)
            if name in self.bin_includes:
                self.bin_includes.remove(name)
                self.bin_excludes.append(name)

        target_path = self.targetdir / exe.target_name
        self._copy_file(
            exe.base,
            target_path,
            copy_dependent_files=False,
            include_mode=True,
        )
        if not os.access(target_path, os.W_OK):
            mode = target_path.stat().st_mode
            target_path.chmod(mode | stat.S_IWUSR)

        # Add resources like version metadata and icon
        self._add_resources(exe)

    def _platform_add_extra_dependencies(
        self, dependent_files: Set[Path]
    ) -> None:
        """Override with platform specific files to add runtime libraries to
        the list of dependent_files calculated in _freeze_executable."""

    @abstractmethod
    def _copy_top_dependency(self, source: Path) -> None:
        """Called for copying the top dependencies in _freeze_executable."""

    def _default_bin_excludes(self) -> List[str]:
        """Return the file names of libraries that need not be included because
        they would normally be expected to be found on the target system or
        because they are part of a package which requires independent
        installation anyway.
        (overridden on Windows)"""
        return ["libclntsh.so", "libwtc9.so", "ldd"]

    def _default_bin_includes(self) -> List[str]:
        """Return the file names of libraries which must be included for the
        frozen executable to work.
        (overriden on Windows)"""
        python_shared_libs = []
        # Miniconda python 3.6-3.9 linux returns a static library to indicate
        # the usage of libpython-static (a shared library is not used).
        name = sysconfig.get_config_var("INSTSONAME")
        if name and not name.endswith(".a"):
            python_shared_libs.append(self._remove_version_numbers(name))
        return python_shared_libs

    @abstractmethod
    def _default_bin_path_excludes(self) -> List[str]:
        """Return the paths of directories which contain files that should not
        be included, generally because they contain standard system
        libraries."""

    @abstractmethod
    def _default_bin_path_includes(self) -> List[str]:
        """Return the paths of directories which contain files that should
        be included."""

    def _get_module_finder(self) -> ModuleFinder:
        finder = ModuleFinder(
            self.include_files,
            self.excludes,
            self.path,
            self.replacePaths,
            self.zipIncludeAllPackages,
            self.zipExcludePackages,
            self.zipIncludePackages,
            self.constants_module,
            self.zip_includes,
        )
        finder.optimize_flag = self.optimize_flag
        for name in self.includes:
            finder.include_module(name)
        for name in self.packages:
            finder.include_package(name)
        return finder

    def _post_freeze_hook(self) -> None:
        """Platform-specific post-Freeze work."""

    def _print_report(self, filename: Path, modules: List[Module]) -> None:
        print(f"writing zip file {filename}\n")
        print("  {:<25} {}".format("Name", "File"))
        print("  {:<25} {}".format("----", "----"))
        for module in modules:
            if module.path:
                print("P", end="")
            else:
                print("m", end="")
            print(f" {module.name:<25} {module.file or ''}")

    @staticmethod
    def _remove_version_numbers(filename: str) -> str:
        tweaked = False
        parts = filename.split(".")
        while parts:
            if not parts[-1].isdigit():
                break
            parts.pop(-1)
            tweaked = True
        if tweaked:
            filename = ".".join(parts)
        return filename

    def _should_copy_file(self, path: Path) -> bool:
        """Return true if the file should be copied to the target machine.
        This is done by checking the bin_path_includes, bin_path_excludes,
        bin_includes and bin_excludes configuration variables using first
        the full file name, then just the base file name, then the file name
        without any version numbers.

        Files are included unless specifically excluded but inclusions take
        precedence over exclusions."""

        # check the full path
        if path in self.bin_includes:
            return True
        if path in self.bin_excludes:
            return False

        # check the file name by itself (with any included version numbers)
        filename = Path(path.name)
        if filename in self.bin_includes:
            return True
        if filename in self.bin_excludes:
            return False

        # check the file name by itself (version numbers removed)
        filename = Path(self._remove_version_numbers(path.name))
        if filename in self.bin_includes:
            return True
        if filename in self.bin_excludes:
            return False

        # check the path for inclusion/exclusion
        dirname = path.parent
        for binpath in self.bin_path_includes:
            try:
                dirname.relative_to(binpath)
            except ValueError:
                pass
            else:
                return True
        for binpath in self.bin_path_excludes:
            try:
                dirname.relative_to(binpath)
            except ValueError:
                pass
            else:
                return False

        return True

    def _verify_configuration(self) -> None:
        """Verify and normalize names and paths. Raises ConfigError on
        failure."""
        filenames = list(self.bin_includes or [])
        filenames += self._default_bin_includes()
        self.bin_includes = [Path(name) for name in filenames]

        filenames = list(self.bin_excludes or [])
        filenames += self._default_bin_excludes()
        self.bin_excludes = [Path(name) for name in filenames]

        paths = list(self.bin_path_includes or [])
        paths += self._default_bin_path_includes()
        self.bin_path_includes = [
            name for name in paths if Path(name).is_dir()
        ]

        paths = list(self.bin_path_excludes or [])
        paths += self._default_bin_path_excludes()
        self.bin_path_excludes = [
            name for name in paths if Path(name).is_dir()
        ]

        if self.zipIncludePackages is None and self.zipExcludePackages is None:
            self.zipIncludePackages = []
            self.zipExcludePackages = ["*"]
        else:
            self.zipIncludePackages = list(self.zipIncludePackages or [])
            self.zipExcludePackages = list(self.zipExcludePackages or [])
        self.zipExcludeAllPackages = "*" in self.zipExcludePackages
        self.zipIncludeAllPackages = "*" in self.zipIncludePackages
        if self.zipExcludeAllPackages and self.zipIncludeAllPackages:
            raise ConfigError(
                "all packages cannot be included and excluded "
                "from the zip file at the same time"
            )
        for name in self.zipIncludePackages:
            if name in self.zipExcludePackages:
                raise ConfigError(
                    f"package {name!r} cannot be both included and "
                    "excluded from zip file"
                )

    def _write_modules(self, filename: Path, finder: ModuleFinder):
        finder.include_file_as_module(
            *self.constants_module.create(finder.modules)
        )

        modules = [m for m in finder.modules if m.name not in finder.excludes]
        modules.sort(key=lambda m: m.name)

        if self.silent < 1:
            self._print_report(filename, modules)
        if self.silent < 2:
            finder.report_missing_modules()

        target_lib_dir = filename.parent
        self._create_directory(target_lib_dir)

        # Prepare zip file
        compress_type = ZIP_DEFLATED if self.compress else ZIP_STORED
        with PyZipFile(filename, "w", compress_type) as outfile:

            files_to_copy: List[Tuple[Module, Path]] = []
            ignorePatterns = shutil.ignore_patterns(
                "*.py", "*.pyc", "*.pyo", "__pycache__"
            )

            for module in modules:

                # determine if the module should be written to the file system;
                # a number of packages make the assumption that files that they
                # require will be found in a location relative to where they
                # are located on disk; these packages will fail with strange
                # errors when they are written to a zip file instead
                include_in_file_system = module.in_file_system
                mod_name = module.name
                mod_name_parts = mod_name.split(".")

                # if the module refers to a package, check to see if this
                # package should be written to the file system
                if (
                    include_in_file_system >= 1
                    and module.path is not None
                    and module.file is not None
                ):
                    parts = mod_name_parts
                    target_package_dir = target_lib_dir.joinpath(*parts)
                    if include_in_file_system == 2:
                        # a few packages are optimized on the hooks,
                        # so for now create the directory for this package
                        self._create_directory(target_package_dir)

                    elif not target_package_dir.exists():
                        # whether the package and its data will be written to
                        # the file system, any non-Python files are copied at
                        # this point if the target directory does not already
                        # exist
                        source_package_dir = module.file.parent
                        if self.silent < 1:
                            print(f"copying data from package {mod_name}...")
                        shutil.copytree(
                            source_package_dir,
                            target_package_dir,
                            ignore=ignorePatterns,
                        )

                        # remove the subfolders which belong to excluded
                        # modules
                        excluded_folders = [
                            m[len(mod_name) + 1 :].replace(".", os.sep)
                            for m in finder.excludes
                            if m.split(".")[0] == parts[0]
                        ]
                        for folder in excluded_folders:
                            folder_to_remove = target_package_dir / folder
                            if folder_to_remove.is_dir():
                                if self.silent < 1:
                                    print(f"removing {folder_to_remove}...")
                                shutil.rmtree(folder_to_remove)

                # if an extension module is found in a package that is to be
                # included in a zip file, copy the actual file to the build
                # directory because shared libraries cannot be loaded from a
                # zip file
                if (
                    module.code is None
                    and module.file is not None
                    and include_in_file_system == 0
                ):
                    parts = mod_name_parts[:-1]
                    parts.append(module.file.name)
                    target = target_lib_dir / ".".join(parts)
                    files_to_copy.append((module, target))

                # starting with Python 3.3 the pyc file format contains the
                # source size; it is not actually used for anything except
                # determining if the file is up to date so we can safely set
                # this value to zero
                if module.code is not None:
                    if module.file is not None and module.file.exists():
                        st = module.file.stat()
                        mtime = int(st.st_mtime)
                        size = st.st_size & 0xFFFFFFFF
                    else:
                        mtime = int(time.time())
                        size = 0
                    header = MAGIC_NUMBER + struct.pack("<iii", 0, mtime, size)
                    if sys.version_info[:2] < (3, 7):
                        header = MAGIC_NUMBER + struct.pack("<ii", mtime, size)
                    data = header + marshal.dumps(module.code)

                # if the module should be written to the file system, do so
                if include_in_file_system >= 1 and module.file is not None:
                    parts = mod_name_parts.copy()
                    if module.code is None:
                        parts.pop()
                        parts.append(module.file.name)
                        target_name = target_lib_dir.joinpath(*parts)
                        self._copy_file(
                            module.file,
                            target_name,
                            copy_dependent_files=True,
                        )
                    else:
                        if module.path is not None:
                            parts.append("__init__")
                        target_name = target_lib_dir.joinpath(*parts)
                        target_name = target_name.with_suffix(".pyc")
                        target_name.write_bytes(data)

                # otherwise, write to the zip file
                elif module.code is not None:
                    zipTime = time.localtime(mtime)[:6]
                    target_name = "/".join(mod_name_parts)
                    if module.path:
                        target_name += "/__init__"
                    zinfo = ZipInfo(target_name + ".pyc", zipTime)
                    zinfo.compress_type = compress_type
                    outfile.writestr(zinfo, data)

            # put the distribution files metadata in the zip file
            dist_cachedir = None
            for module in modules:
                if module.distribution:
                    dist_cachedir = module.distribution.locate_file(".")
                    break
            if dist_cachedir is not None:
                pos = len(dist_cachedir.as_posix()) + 1
                for name in dist_cachedir.rglob("*"):
                    if name.is_dir():
                        continue
                    outfile.write(name, name.as_posix()[pos:])

            # write any files to the zip file that were requested specially
            for source_path, target_path in finder.zip_includes:
                if source_path.is_dir():
                    pos = len(source_path.as_posix()) + 1
                    for source_filename in source_path.rglob("*"):
                        if source_filename.is_dir():
                            continue
                        target = target_path / source_filename.as_posix()[pos:]
                        outfile.write(source_filename, target)
                else:
                    outfile.write(source_path, target_path.as_posix())

        # Copy Python extension modules from the list built above.
        origPath = os.environ["PATH"]
        for module, target in files_to_copy:
            try:
                if module.parent is not None:
                    path = os.pathsep.join(
                        [origPath] + [str(p) for p in module.parent.path]
                    )
                    os.environ["PATH"] = path
                self._copy_file(module.file, target, copy_dependent_files=True)
            finally:
                os.environ["PATH"] = origPath

    def Freeze(self):
        self.files_copied: Set[Path] = set()

        finder: ModuleFinder = self._get_module_finder()
        self.finder: ModuleFinder = finder

        # Add the executables to target
        for executable in self.executables:
            self._freeze_executable(executable)

        # Write the modules
        targetdir = self.targetdir
        library_zip = targetdir / "lib" / "library.zip"
        self._write_modules(library_zip, finder)

        excluded_dependent_files = self.finder.excluded_dependent_files
        for source_path, target_path in finder.included_files:
            copy_dependent_files = source_path not in excluded_dependent_files
            if source_path.is_dir():
                # Copy directories by recursing into them.
                # Can't use shutil.copytree because we may need dependencies
                target_base = targetdir / target_path
                for name in source_path.rglob("*"):
                    if name.is_dir():
                        continue
                    if ".svn" in name.parents:
                        continue
                    if "CVS" in name.parents:
                        continue
                    fulltarget = target_base / name.relative_to(source_path)
                    self._create_directory(fulltarget.parent)
                    self._copy_file(name, fulltarget, copy_dependent_files)
            else:
                # Copy regular files.
                fulltarget = targetdir / target_path
                self._copy_file(source_path, fulltarget, copy_dependent_files)

        # do any platform-specific post-Freeze work
        self._post_freeze_hook()


class WinFreezer(Freezer, PEParser):
    def __init__(self, *args, **kwargs):
        Freezer.__init__(self, *args, **kwargs)
        PEParser.__init__(self, self.silent)

        # deal with C-runtime files
        self.runtime_files: Set[str] = set()
        self.runtime_files_to_dup: Set[str] = set()
        self._set_runtime_files()

    def _add_resources(self, exe: Executable) -> None:
        target_path: Path = self.targetdir / exe.target_name

        # Change the manifest
        manifest: Optional[str] = exe.manifest
        if manifest is not None or exe.uac_admin:
            if self.silent < 1:
                print(f"writing manifest -> {target_path}")
            try:
                if exe.uac_admin:
                    manifest = manifest or self.read_manifest(target_path)
                    manifest = manifest.replace(
                        "asInvoker", "requireAdministrator"
                    )
                self.write_manifest(target_path, manifest)
            except FileNotFoundError as exc:
                if self.silent < 3:
                    print("WARNING:", exc)
            except RuntimeError as exc:
                if self.silent < 3:
                    print(f"WARNING: error parsing {target_path}:", exc)

        # Add version resource
        if self.metadata is not None:
            warning_msg = "WARNING: unable to create version resource:"
            if not self.metadata.version:
                if self.silent < 3:
                    print(warning_msg, "version must be specified")
            else:
                version = VersionInfo(
                    self.metadata.version,
                    comments=self.metadata.long_description,
                    description=self.metadata.description,
                    company=self.metadata.author,
                    product=self.metadata.name,
                    copyright=exe.copyright,
                    trademarks=exe.trademarks,
                    verbose=bool(self.silent < 1),
                )
                try:
                    version.stamp(target_path)
                except (FileNotFoundError, RuntimeError) as exc:
                    if self.silent < 3:
                        print(warning_msg, exc)

        # Add icon
        if exe.icon is not None:
            try:
                AddIcon(target_path, exe.icon)
            except MemoryError:
                if self.silent < 3:
                    print("WARNING: MemoryError")
            except RuntimeError as exc:
                if self.silent < 3:
                    print("WARNING:", exc)
            except OSError as exc:
                if "\\WindowsApps\\" in sys.base_prefix:
                    if self.silent < 3:
                        print("WARNING:", exc)
                        print(
                            "WARNING: Because of restrictions on Microsoft "
                            "Store apps, Python scripts may not have full "
                            "write access to built executable. "
                            "You will need to install the full installer."
                        )
                else:
                    raise

        # Update the PE checksum (or fix it in case it is zero)
        try:
            UpdateCheckSum(target_path)
        except MemoryError:
            if self.silent < 3:
                print("WARNING: MemoryError")
        except (RuntimeError, OSError) as exc:
            if self.silent < 3:
                print("WARNING:", exc)

    def _copy_top_dependency(self, source: Path) -> None:
        """Called for copying certain top dependencies in _freeze_executable.
        We need this as a separate method so that it can be overridden on
        Darwin and Windows."""
        # top dependencies do not go into lib on windows, but the C runtimes
        # can be handled in _copy_file/_pre_copy_hook
        target = self.targetdir / source.name
        self._copy_file(
            source, target, copy_dependent_files=True, include_mode=True
        )

    def _pre_copy_hook(self, source: Path, target: Path) -> Tuple[Path, Path]:
        """Prepare the source and target paths. Also, adjust the target of
        C runtime libraries and triggers an additional copy for files in
        self.runtime_files_to_dup."""

        # fix the target path for C runtime files
        norm_target_name = target.name.lower()
        if norm_target_name in self.runtime_files:
            target_name = target.name
            target = self.targetdir / "lib" / target_name

            # vcruntime140.dll must be in the root and in the lib directory
            if norm_target_name in self.runtime_files_to_dup:
                self.runtime_files_to_dup.remove(norm_target_name)
                self._copy_file(source, target, copy_dependent_files=False)
                target = self.targetdir / target_name
        return source, target

    def _post_copy_hook(
        self,
        source: Path,
        target: Path,
        copy_dependent_files: bool,
        include_mode: bool = False,
    ) -> None:
        if (
            copy_dependent_files
            and source not in self.finder.excluded_dependent_files
        ):

            targetdir = target.parent
            for dependent_file in self.get_dependent_files(source):
                target = targetdir / dependent_file.name
                self._copy_file(dependent_file, target, copy_dependent_files)

    def _default_bin_excludes(self) -> List[str]:
        return ["comctl32.dll", "oci.dll"]

    def _default_bin_includes(self) -> List[str]:
        python_shared_libs: List[str] = []
        if MINGW:
            name = sysconfig.get_config_var("INSTSONAME")
            if name:
                python_shared_libs.append(name.replace(".dll.a", ".dll"))
        else:
            python_shared_libs += [
                "python%s.dll" % sys.version_info[0],
                "python%s%s.dll" % sys.version_info[:2],
            ]
        return python_shared_libs

    def _default_bin_path_excludes(self) -> List[str]:
        systemDir = GetSystemDir()
        windowsDir = GetWindowsDir()
        return [windowsDir, systemDir, os.path.join(windowsDir, "WinSxS")]

    def _default_bin_path_includes(self) -> List[str]:
        paths = {Path(path) for path in sys.path if path}
        # force some paths for conda systems
        paths.add(Path(sys.prefix, "DLLs"))
        paths.add(Path(sys.prefix, "Library", "bin"))
        # do the same for msys2, mingw32/64
        if sysconfig.get_config_var("DESTSHARED"):
            paths.add(Path(sysconfig.get_config_var("DESTSHARED")))
        # return only valid paths
        return [str(path) for path in paths if path.is_dir()]

    def _platform_add_extra_dependencies(
        self, dependent_files: Set[Path]
    ) -> None:
        search_dirs: Set[Path] = set()
        for filename in dependent_files:
            search_dirs.add(filename.parent)
        for filename in self.runtime_files:
            for search_dir in search_dirs:
                filepath = search_dir / filename
                if filepath.exists():
                    dependent_files.add(filepath)

    def _post_freeze_hook(self) -> None:
        target_lib = self.targetdir / "lib"
        # Recursing into directories to search for load order files.
        # Some libraries use delvewheel to patch them.
        for loader_file in target_lib.rglob(".load-order-*"):
            load_order = loader_file.read_text().split()
            load_dir = loader_file.parent
            new_order = [
                f for f in load_order if load_dir.joinpath(f).is_file()
            ]
            if new_order != load_order:
                loader_file.write_text("\n".join(new_order))

    def _set_runtime_files(self) -> None:
        if self.include_msvcr:
            self.runtime_files.update(winmsvcr.FILES)
            self.runtime_files_to_dup.update(winmsvcr.FILES_TO_DUPLICATE)
        else:
            # just put on the exclusion list
            self.bin_excludes.extend([Path(name) for name in winmsvcr.FILES])


class DarwinFreezer(Freezer, Parser):
    def __init__(self, *args, **kwargs):
        Freezer.__init__(self, *args, **kwargs)
        Parser.__init__(self, self.silent)
        self.darwin_tracker: Optional[DarwinFileTracker] = None
        self.darwin_tracker = DarwinFileTracker()

    def _post_freeze_hook(self) -> None:
        self.darwin_tracker.finalizeReferences()

    def _pre_copy_hook(self, source: Path, target: Path) -> Tuple[Path, Path]:
        """Prepare the source and target paths."""
        return source, target

    def _post_copy_hook(
        self,
        source: Path,
        target: Path,
        copy_dependent_files: bool,
        include_mode: bool = False,
        reference: Optional["MachOReference"] = None,
    ) -> None:

        # The file was not previously copied, so need to create a
        # DarwinFile file object to represent the file being copied.
        if reference is not None:
            referencing_file = reference.source_file
        else:
            referencing_file = None
        darwin_file = DarwinFile(source, referencing_file)
        darwin_file.setBuildPath(target)
        if reference is not None:
            reference.setTargetFile(darwin_file)

        self.darwin_tracker.recordCopiedFile(target, darwin_file)
        if (
            copy_dependent_files
            and source not in self.finder.excluded_dependent_files
        ):
            # Always copy dependent files on root directory
            # to allow to set relative reference
            targetdir = self.targetdir
            for dependent in self.get_dependent_files(source, darwin_file):
                target = targetdir / dependent.name
                reference = darwin_file.getMachOReferenceForPath(dependent)
                self._copy_file_recursion(
                    dependent,
                    target,
                    copy_dependent_files=True,
                    reference=reference,
                )

    def _copy_file_recursion(
        self,
        source: Path,
        target: Path,
        copy_dependent_files: bool,
        include_mode: bool = False,
        reference: Optional["MachOReference"] = None,
    ) -> None:
        """This is essentially the same as Freezer._copy_file, except that it
        also takes a reference parameter. Used when recursing to dependencies
        of a file on Darwin."""
        if not self._should_copy_file(source):
            return

        # handle pre-copy tasks, normally on the target path
        source, target = self._pre_copy_hook(source, target)

        if target in self.files_copied:
            if reference is not None:
                # If file was already copied, and we are following a reference
                # from a DarwinFile, then we need to tell the reference where
                # the file was copied to (the reference can later be updated).
                reference.setTargetFile(
                    self.darwin_tracker.getDarwinFile(source, target)
                )
            return
        if source == target:
            return
        self._create_directory(target.parent)
        if self.silent < 1:
            print(f"copying {source} -> {target}")
        shutil.copyfile(source, target)
        shutil.copystat(source, target)
        if include_mode:
            shutil.copymode(source, target)
        self.files_copied.add(target)

        # handle post-copy tasks, including copying dependencies
        self._post_copy_hook(
            source,
            target,
            copy_dependent_files=copy_dependent_files,
            include_mode=include_mode,
            reference=reference,
        )

    def _copy_top_dependency(self, source: Path) -> None:
        """Called for copying certain top dependencies. We need this as a
        separate function so that it can be overridden on Darwin
        (to interact with the DarwinTools system)."""

        target = self.targetdir / "lib" / source.name

        # this recovers the cached MachOReference pointers to the files
        # found by the get_dependent_files calls made previously (if any).
        # If one is found, pass into _copy_file.
        # We need to do this so the file knows what file referenced it,
        # and can therefore calculate the appropriate rpath.
        # (We only cache one reference.)
        cachedReference = self.darwin_tracker.getCachedReferenceTo(source)
        self._copy_file_recursion(
            source,
            target,
            copy_dependent_files=True,
            include_mode=True,
            reference=cachedReference,
        )

    def _default_bin_path_excludes(self):
        return ["/lib", "/usr/lib", "/System/Library/Frameworks"]

    def _default_bin_path_includes(self) -> List[str]:
        return [sysconfig.get_config_var("DESTSHARED")]

    def get_dependent_files(
        self, path: Path, darwinFile: Optional["DarwinFile"] = None
    ) -> Set[Path]:
        try:
            return self.dependent_files[path]
        except KeyError:
            pass

        # if darwinFile is None (which means that get_dependent_files is
        # being called outside of _copy_file -- e.g., one of the
        # preliminary calls in _freeze_executable), create a temporary
        # DarwinFile object for the path, just so we can read its
        # dependencies
        if darwinFile is None:
            darwinFile = DarwinFile(path)
        dependent_files = darwinFile.getDependentFilePaths()

        # cache the MachOReferences to the dependencies, so they can be
        # called up later in _copy_file if copying a dependency without
        # an explicit reference provided (to assist in resolving @rpaths)
        for reference in darwinFile.getMachOReferenceList():
            if reference.isResolved():
                self.darwin_tracker.cacheReferenceTo(
                    reference.resolved_path, reference
                )
        self.dependent_files[path] = dependent_files
        return dependent_files


class LinuxFreezer(Freezer, ELFParser):
    def __init__(self, *args, **kwargs):
        Freezer.__init__(self, *args, **kwargs)
        ELFParser.__init__(self, self.silent)
        self._symlinks: Set[Tuple[Path, str]] = set()

    def _pre_copy_hook(self, source: Path, target: Path) -> Tuple[Path, Path]:
        """Prepare the source and target paths. In addition, it ensures that
        the source of a symbolic link is copied by deferring the creation of
        the link."""
        if source.is_symlink():
            real_source = source.resolve()
            symlink = real_source.name
            real_target = target.with_name(symlink)
            self._symlinks.add((target, symlink))
            return real_source, real_target
        return source, target

    def _post_copy_hook(
        self,
        source: Path,
        target: Path,
        copy_dependent_files: bool,
        include_mode: bool = False,
    ):
        if (
            copy_dependent_files
            and source not in self.finder.excluded_dependent_files
        ):
            targetdir = target.parent
            source_dir = source.parent
            library_dir = self.targetdir / "lib"
            fix_rpath = set()
            for dependent_file in self.get_dependent_files(source):
                if not self._should_copy_file(dependent_file):
                    continue
                try:
                    relative = dependent_file.relative_to(source_dir)
                except ValueError:
                    # put it in the targetdir if not already copied
                    dependent_filename = dependent_file.name
                    dependent_target = targetdir / dependent_filename
                    relative = Path(dependent_filename)
                    if dependent_target not in self.files_copied:
                        for file in self.files_copied:
                            if file.name == dependent_filename:
                                relative = file.relative_to(library_dir)
                                dependent_target = library_dir / relative
                                break
                else:
                    dependent_target = targetdir / relative
                    dependent_target = dependent_target.resolve()
                    try:
                        dependent_target.relative_to(library_dir)
                    except ValueError:
                        parts = list(relative.parts)
                        while parts[0] == os.pardir:
                            parts.pop(0)
                        relative = Path(*parts)
                        dependent_target = library_dir / relative
                dep_libs = str(relative.parent)
                fix_rpath.add(dep_libs)
                self._copy_file(
                    dependent_file, dependent_target, copy_dependent_files
                )
            if fix_rpath:
                has_rpath = self.get_rpath(target)
                rpath = ":".join(f"$ORIGIN/{r}" for r in fix_rpath)
                if has_rpath != rpath:
                    self.set_rpath(target, rpath)

    def _post_freeze_hook(self):
        target: Path
        symlink: str
        for target, symlink in self._symlinks:
            if self.silent < 1:
                print(f"linking {target} -> {symlink}")
            if not target.exists():
                target.symlink_to(symlink)

    def _copy_top_dependency(self, source: Path) -> None:
        """Called for copying the top dependencies in _freeze_executable."""
        target = self.targetdir / "lib" / source.name
        self._copy_file(
            source, target, copy_dependent_files=True, include_mode=True
        )

    def _default_bin_path_excludes(self) -> List[str]:
        return [
            "/lib",
            "/lib32",
            "/lib64",
            "/usr/lib",
            "/usr/lib32",
            "/usr/lib64",
        ]

    def _default_bin_path_includes(self) -> List[str]:
        # add the stdlib/lib-dynload directory
        destlib = sysconfig.get_config_var("DESTLIB")
        if bool(destlib):
            return [destlib]
        return []
